"""Tests for the `ScourProtectionDesign` class."""

__author__ = "Rob Hammond"
__copyright__ = "Copyright 2020, National Renewable Energy Laboratory"
__maintainer__ = "Rob Hammond"
__email__ = "robert.hammond@nrel.gov"


from copy import deepcopy

import numpy as np
import pytest

from wisdem.orbit.phases.design import ScourProtectionDesign

config_min_defined = {
    "monopile": {"diameter": 9},
    "plant": {"num_turbines": 50},
    "scour_protection_design": {"cost_per_tonne": 400},
}

config_fully_defined = {
    "monopile": {"diameter": 10},
    "plant": {"num_turbines": 50},
    "scour_protection_design": {
        "rock_density": 2300,
        "cost_per_tonne": 400,
        "design_time": 500,
        "soil_friction_angle": 33.0,
        "scour_depth_equilibrium": 1.2,
        "scour_protection_depth": 0.3,
    },
}


def test_default_setup():
    scour = ScourProtectionDesign(config_min_defined)
    scour.run()

    assert scour.phi == 33.5
    assert scour.equilibrium == 1.3
    assert scour.rock_density == 2600


def test_fully_defined_setup():
    scour = ScourProtectionDesign(config_fully_defined)
    scour.run()

    design = config_fully_defined["scour_protection_design"]

    assert scour.phi == design["soil_friction_angle"]
    assert scour.equilibrium == design["scour_depth_equilibrium"]
    assert scour.rock_density == design["rock_density"]


@pytest.mark.parametrize(
    "config,expected",
    ((config_fully_defined, 1141), (config_min_defined, 3852)),
    ids=["fully_defined", "minimally_defined"],
)
def test_tonnes_per_substructure(config, expected):
    scour = ScourProtectionDesign(config)
    scour.run()

    assert scour.scour_protection_tonnes == expected


@pytest.mark.parametrize(
    "config",
    (config_fully_defined, config_min_defined),
    ids=["fully_defined", "minimally_defined"],
)
def test_total_cost(config):
    scour = ScourProtectionDesign(config_min_defined)
    scour.run()

    cost = (
        config["plant"]["num_turbines"]
        * config["scour_protection_design"]["cost_per_tonne"]
        * scour.scour_protection_tonnes
    )
    assert scour.total_cost == pytest.approx(cost, rel=1e-8)



import numpy as np
from numpy.linalg import norm
from scipy.interpolate import griddata


def project_points(points, x, N_vect):
    """
    Rotate the surface and the points in order to align the vector N_vect in the z direction
    """

    N_vect = N_vect / norm(N_vect)
    rot = calculate_rotation_matrix(N_vect)
    points_rot = dotX(rot, points)
    x_rot = dotX(rot, x)
    points_rot2 = np.array(
        [
            points_rot[:, 0],
            points_rot[:, 1],
            griddata(
                (x_rot[:, :, 0].flatten(), x_rot[:, :, 1].flatten()),
                x_rot[:, :, 2].flatten(),
                (points_rot[:, 0], points_rot[:, 1]),
                method="linear",
            ),
        ]
    ).T
    ### Rotate back
    inv_rot = np.linalg.inv(rot)
    points_final = dotX(inv_rot, points_rot2)
    return points_final


def normalize(v):
    return v / (np.dot(v, v) ** 2 + 1.0e-16)


def curvature(points):
    if len(points.shape) < 2:
        return None
    if points.shape[1] == 1:
        return None

    if points.shape[1] == 2:
        d1 = np.diff(points.T)
        d2 = np.diff(d1)
        x1 = d1[0, 1:]
        y1 = d1[1, 1:]
        x2 = d2[0, :]
        y2 = d2[1, :]
        curv = (x1 * y2 - y1 * x2) / (x1**2 + y1**2) ** (3.0 / 2.0)

    elif points.shape[1] == 3:
        d1 = np.diff(points.T)
        d2 = np.diff(d1)
        x1 = d1[0, 1:]
        y1 = d1[1, 1:]
        z1 = d1[2, 1:]
        x2 = d2[0, :]
        y2 = d2[1, :]
        z2 = d2[2, :]
        curv = ((z2 * y1 - y2 * z1) ** 2.0 + (x2 * z1 - z2 * x1) ** 2.0 + (y2 * x1 - x2 * y1) ** 2.0) ** 0.5 / (
            x1**2.0 + y1**2.0 + z1**2.0 + 1.0e-30
        ) ** (3.0 / 2.0)

    curvt = np.zeros(points.shape[0])
    try:
        curvt[1:-1] = curv
        curvt[0] = curv[0]
        curvt[-1] = curv[-1]
    except:
        pass
    return curvt


def calculate_angle(v1, v2):
    """
    Calculate the signed angle between the vector, v1 and the vector, v2

    \param    v1      <c>array(3)</c>   :       vector 1
    \param    v2      <c>array(3)</c>   :       vector 2
    \retval   angle   <c>float radian</c> :     the angle in the two vector plane
    """

    v1 = v1 / norm(v1)
    v2 = v2 / norm(v2)
    return np.arctan2(norm(np.cross(v1, v2)), np.dot(v1, v2))


def calculate_rotation_matrix(vect):
    """
    Transpose (P1) and project the normal vector (vect) to the Z direction.

    \param    vect        vector to rotate                <c>numpy.array(3)</c>
    \retval   array       the rotation matrix             <c>numpy.array((3,3))</c>
    """
    return rotation_matrix_global(vect, np.array([0, 0, 1]))


def rotation_matrix_global(vect, direction):
    """
    Transpose (P1) and project the normal vector (vect) to the direction.

    \param    vect        vector to rotate               <c>numpy.array(3)</c>
    \param    direction   direction to rotate towards    <c>numpy.array(3)</c>
    \retval   array       the rotation matrix            <c>numpy.array((3,3))</c>
    """
    ### Normalize vect
    vect_norm = vect / norm(vect)
    ### Calculate the vector normal to the normal vector and the direction
    w = np.cross(vect_norm, direction)
    if norm(w) == 0:
        ### The two vectors are coplanar, exit with an identity matrix
        if np.dot(vect_norm, direction) > 0:
            return np.array([[1, 0, 0], [0, 1, 0], [0, 0, 1]])
        else:
            return -np.array([[1, 0, 0], [0, 1, 0], [0, 0, 1]])

    w_norm = w / norm(w)
    # print 'w_norm',w_norm
    ### The angle is found by taking the arccos of vect_norm[2]
    # q = math.acos(vect_norm[2])
    q = calculate_angle(vect_norm, direction)
    # print 'q',q
    ### calculate the rotation matrix of the vector w_norm and angle q
    rot = RotMat(w_norm, q)
    return rot


def dotX(rot, x, trans_vect=np.array([0.0, 0.0, 0.0])):
    """
    Transpose and Multiply the x array by a rotational matrix
    """
    if isinstance(x, list):
        x_tmp = np.array([x[0].flatten(), x[1].flatten(), x[2].flatten()]).T
        x_rot_tmp = np.zeros(x_tmp.shape)
        for i in range(x_tmp.shape[0]):
            x_rot_tmp[i, :] = dot(rot, x_tmp[i, :] - trans_vect)

        x_rot = []
        for iX in range(3):
            x_rot.append(x_rot_tmp[:, iX].reshape(x[0].shape))
    elif isinstance(x, np.ndarray):
        x_rot = np.zeros(x.shape)
        if len(x.shape) == 1:
            x_rot = np.dot(rot, x - trans_vect)
        elif len(x.shape) == 2:
            for i in range(x.shape[0]):
                x_rot[i, :] = np.dot(rot, x[i, :] - trans_vect)
        elif len(x.shape) == 3:
            for i in range(x.shape[0]):
                for j in range(x.shape[1]):
                    x_rot[i, j, :] = np.dot(rot, x[i, j, :] - trans_vect)

    return x_rot


# rotation matrix function for an x-rotation
RotX = lambda a: np.array([[1, 0, 0], [0, np.cos(a), -np.sin(a)], [0, np.sin(a), np.cos(a)]])
# rotation matrix function for a y-rotation
RotY = lambda a: np.array([[np.cos(a), 0, np.sin(a)], [0, 1, 0], [-np.sin(a), 0, np.cos(a)]])
# rotation matrix function for a z-rotation
RotZ = lambda a: np.array([[np.cos(a), -np.sin(a), 0], [np.sin(a), np.cos(a), 0], [0, 0, 1]])


def RotMat(u, theta):
    """
    Calculate the rotation matrix from a unit vector, u and an angle, theta

    http://en.wikipedia.org/wiki/Rotation_matrix#Rotation_matrix_from_axis_and_angle
    \f[
        R =\left[ \begin{array}{ccc}
        \cos\theta +u_x^2 \left(1-\cos \theta\right) &
        u_x u_y \left(1-\cos \theta\right) - u_z \sin \theta &
        u_x u_z \left(1-\cos \theta\right) + u_y \sin \theta \\
        u_y u_x \left(1-\cos \theta\right) + u_z \sin \theta &
        \cos \theta + u_y^2\left(1-\cos \theta\right) &
        u_y u_z \left(1-\cos \theta\right) - u_x \sin \theta \\
        u_z u_x \left(1-\cos \theta\right) - u_y \sin \theta &
        u_z u_y \left(1-\cos \theta\right) + u_x \sin \theta &
        \cos \theta + u_z^2\left(1-\cos \theta\right)
        \end{array} \right]
    \f]

    \param    u       <c> list/tuple/array(3) </c>       vector of the direction to rotate from
    \param    theta    <c> int radian </c>               angle to rotate with

    \retval   array   <c> array(3,3) </c>                rotation matrix
    """
    from numpy import cos, sin, array

    rot = array(
        [
            [
                cos(theta) + u[0] ** 2 * (1 - cos(theta)),
                u[0] * u[1] * (1 - cos(theta)) - u[2] * sin(theta),
                u[0] * u[2] * (1 - cos(theta)) + u[1] * sin(theta),
            ],
            [
                u[1] * u[0] * (1 - cos(theta)) + u[2] * sin(theta),
                cos(theta) + u[1] ** 2 * (1 - cos(theta)),
                u[1] * u[2] * (1 - cos(theta)) - u[0] * sin(theta),
            ],
            [
                u[2] * u[0] * (1 - cos(theta)) - u[1] * sin(theta),
                u[2] * u[1] * (1 - cos(theta)) + u[0] * sin(theta),
                cos(theta) + u[2] ** 2 * (1 - cos(theta)),
            ],
        ]
    )
    return rot


def dotXC(rot, x, center):
    """
    Transpose and Multiply the x array by a rotational matrix around a center
    """
    from numpy import dot, array, zeros

    if isinstance(x, list):
        x_tmp = array([x[0].flatten(), x[1].flatten(), x[2].flatten()]).T
        x_rot_tmp = zeros(x_tmp.shape)
        for i in range(x_tmp.shape[0]):
            x_rot_tmp[i, :] = dot(rot, x_tmp[i, :] - center) + center

        x_rot = []
        for iX in range(3):
            x_rot.append(x_rot_tmp[:, iX].reshape(x[0].shape))
    elif isinstance(x, np.ndarray):
        x_rot = zeros(x.shape)
        if len(x.shape) == 1:
            x_rot[:] = dot(rot, x - center) + center

        if len(x.shape) == 2:
            for i in range(x.shape[0]):
                x_rot[i, :] = dot(rot, x[i, :] - center) + center
        elif len(x.shape) == 3:
            for i in range(x.shape[0]):
                for j in range(x.shape[1]):
                    x_rot[i, j, :] = dot(rot, x[i, j, :] - center) + center
    return x_rot



"""
Testing framework for common oss installation tasks.
"""

__author__ = "Jake Nunemaker"
__copyright__ = "Copyright 2020, National Renewable Energy Laboratory"
__maintainer__ = "Jake Nunemaker"
__email__ = "Jake.Nunemaker@nrel.gov"


import pytest

from wisdem.orbit.core.exceptions import MissingComponent
from wisdem.orbit.phases.install.oss_install.common import lift_topside, attach_topside


@pytest.mark.parametrize(
    "task, log, args",
    [
        (lift_topside, "Lift Topside", []),
        (attach_topside, "Attach Topside", []),
    ],
)
def test_task(env, wtiv, task, log, args):
    env.register(wtiv)
    wtiv.initialize(mobilize=False)

    task(wtiv, *args, site_depth=10)
    env.run()

    actions = [a["action"] for a in env.actions]
    assert log in actions


@pytest.mark.parametrize(
    "task, log, args",
    [
        (lift_topside, "Lift Topside", []),
        (attach_topside, "Attach Topside", []),
    ],
)
def test_task_fails(env, feeder, task, log, args):
    env.register(feeder)
    feeder.initialize(mobilize=False)

    with pytest.raises(MissingComponent):
        task(feeder, *args, site_depth=10)
        env.run()

    actions = [a["action"] for a in env.actions]
    assert log not in actions



"""Tests for common infrastructure for quayside assembly tow-out simulations"""

__author__ = "Jake Nunemaker"
__copyright__ = "Copyright 2020, National Renewable Energy Laboratory"
__maintainer__ = "Jake Nunemaker"
__email__ = "jake.nunemaker@nrel.gov"


import pandas as pd
import pytest

from wisdem.orbit.core import WetStorage
from wisdem.orbit.phases.install.quayside_assembly_tow.common import TurbineAssemblyLine, SubstructureAssemblyLine


@pytest.mark.parametrize(
    "num, assigned, expected",
    [
        (1, [], 0),
        (1, [1] * 10, 100),
        (2, [1] * 10, 50),
        (3, [1] * 10, 40),
        (5, [1] * 10, 20),
        (10, [1] * 10, 10),
    ],
)
def test_SubstructureAssemblyLine(env, num, assigned, expected):
    _assigned = len(assigned)
    storage = WetStorage(env, capacity=float("inf"))

    for a in range(num):
        assembly = SubstructureAssemblyLine(assigned, 10, storage, a + 1)
        env.register(assembly)
        assembly.start()

    env.run()

    assert len(env.actions) == _assigned
    assert env.now == expected


@pytest.mark.parametrize(
    "num, assigned",
    [
        (1, [1] * 10),
        (2, [1] * 10),
        (3, [1] * 10),
        (5, [1] * 10),
        (10, [1] * 10),
    ],
)
def test_TurbineAssemblyLine(env, num, assigned):
    _assigned = len(assigned)
    feed = WetStorage(env, capacity=float("inf"))
    target = WetStorage(env, capacity=float("inf"))

    for i in assigned:
        feed.put(0)

    for a in range(num):
        assembly = TurbineAssemblyLine(feed, target, {"tower": {"sections": 1}}, a + 1)
        env.register(assembly)
        assembly.start()

    env.run()

    df = pd.DataFrame(env.actions)
    assert len(df.loc[df["action"] == "Mechanical Completion"]) == len(assigned)


@pytest.mark.parametrize(
    "sub_lines, turb_lines",
    [
        (1, 1),
        (1, 10),
        (1, 100),
        (10, 1),
        (10, 10),
        (10, 100),
        (100, 1),
        (100, 10),
        (100, 100),
    ],
)
def test_Sub_to_Turbine_assembly_interaction(env, sub_lines, turb_lines):
    num_turbines = 50
    assigned = [1] * num_turbines

    feed = WetStorage(env, capacity=2)
    target = WetStorage(env, capacity=float("inf"))

    for a in range(sub_lines):
        assembly = SubstructureAssemblyLine(assigned, 10, feed, a + 1)
        env.register(assembly)
        assembly.start()

    for a in range(turb_lines):
        assembly = TurbineAssemblyLine(feed, target, {"tower": {"sections": 1}}, a + 1)
        env.register(assembly)
        assembly.start()

    env.run()

    df = pd.DataFrame(env.actions)
    assert len(df.loc[df["action"] == "Mechanical Completion"]) == num_turbines



import os
from wisdem.glue_code.runWISDEM import run_wisdem

## File management
run_dir = os.path.dirname( os.path.realpath(__file__) ) + os.sep
wisdem_examples = os.path.dirname(os.path.dirname(run_dir))
fname_wt_input = os.path.join(wisdem_examples, "03_blade", "BAR_USC.yaml")
fname_modeling_options = os.path.join(wisdem_examples, "03_blade", "modeling_options.yaml")
fname_analysis_options = os.path.join(run_dir, "analysis_options_custom.yaml")

wt_opt, modeling_options, analysis_options = run_wisdem(fname_wt_input, fname_modeling_options, fname_analysis_options)




# 0 ---------- (marker for docs)
import openmdao.api as om

from wisdem.nrelcsm.nrel_csm_mass_2015 import nrel_csm_2015

# 0 ---------- (marker for docs)

# 1 ---------- (marker for docs)
# OpenMDAO Problem instance
prob = om.Problem(reports=False)
prob.model = nrel_csm_2015()
prob.setup()
# 1 ---------- (marker for docs)

# 2 ---------- (marker for docs)
# Initialize variables for NREL CSM
prob["machine_rating"] = 5000.0
prob["rotor_diameter"] = 126.0
prob["turbine_class"] = 2
prob["tower_length"] = 90.0
prob["blade_number"] = 3
prob["blade_has_carbon"] = False
prob["max_tip_speed"] = 80.0
prob["max_efficiency"] = 0.90
prob["main_bearing_number"] = 2
prob["crane"] = True
# 2 ---------- (marker for docs)

# 3 ---------- (marker for docs)
# Evaluate the model
prob.run_model()
# 3 ---------- (marker for docs)

# 4 ---------- (marker for docs)
# Print all intermediate inputs and outputs to the screen
prob.model.list_inputs(units=True)
prob.model.list_outputs(units=True)
# 4 ---------- (marker for docs)



""" 
This module contains:
  - PolarFile: class to read different kind of polar formats

  - 
"""

import os
import numpy as np
import pandas as pd

# --- Welib polar readers
CSVFile=None
ADPolarFile=None
HAWC2AEFile=None

class WrongPolarFormatError(Exception): pass
class BrokenPolarFormatError(Exception): pass


# List of columns used to "unify" the dataframes coming out of "PolarFile"
DEFAULT_COLUMNS={'alpha':'Alpha', 'cl':'Cl', 'cd':'Cd', 'cm':'Cm'}
DEFAULT_COLUMNS_EXT={
        'clinv':'Cl_inv', 'clfs':'Cl_fs', 'fs':'fs',
        'cn':'Cn', 'cnpot':'Cn_pot', 'cnz12':'Cn_012', 'cnf':'Cn_f', 'cncd0off':'Cn_Cd0off'
        }


# --------------------------------------------------------------------------------
# --- Small Helper functions
# --------------------------------------------------------------------------------
def _load_txt(filename, commentChars, skiprows=0, **kwargs):
    """ 
    Similar to np.loadtxt but also works if comments are present anywhere in the file (e.g. end of file)
    """
    with open(filename) as f:
        lines = (line for iline, line in enumerate(f) if not line.startswith(commentChars) and iline>=skiprows)
        Lines = list(lines) 
    if len(Lines)==0:
        raise Exception('Zero lines')
    else:
        return np.loadtxt(Lines, **kwargs)


# --------------------------------------------------------------------------------}
# --- Simple classes 
# --------------------------------------------------------------------------------{
class BasePolarFile(dict):
    def __init__(self, filename=None):
        super().__init__()
        self.COMMENT_CHARS=('#','!','%')
        self['header'] = ''
        self['columns'] = []
        self['data']    = np.array([[]])
        self['nPolars']  = 0
        if filename is not None:
            self.read(filename)

    def __repr__(self):
        s='<{} object>:\n'.format(type(self).__name__)
        s+='- header: {}\n'.format(self['header'])
        s+='- columns: {}\n'.format(self['columns'])
        s+='- nPolars:{}\n'.format(self['nPolars'])
        s+='- data: shape {}\n'.format(self['data'].shape)
        s+='        first: {}\n'.format(self['data'][0,:])
        s+='        last:  {}\n'.format(self['data'][-1,:])
        return s

    @staticmethod
    def formatName(): raise NotImplementedError()

    def toDataFrame(self):
        if self['nPolars']==1:
            return pd.DataFrame(data=self['data'], columns=self['columns'])
        else:
            raise NotImplementedError()

class PolarFile_OneLineHeader(BasePolarFile):
    """ Polar file with exatcly one line of header. 
    Column names in header can be separated by spaces or commas.
    Header may start with the following comment characters: ['#','!','%']
    Data may be space or column separated
    """
    @staticmethod
    def formatName(): return 'Polar file one header line'

    def read(self, filename):
        super().__init__()
        with open(filename) as f:
            header = f.readline().strip()
            second = f.readline()
        self['header'] = header
        for c in self.COMMENT_CHARS:
            header = header.lstrip(c)
        sep=',' 
        try:
            self['data'] = np.loadtxt(filename, delimiter=sep, skiprows=1)
        except:
            sep=None
            self['data'] = np.loadtxt(filename, delimiter=sep, skiprows=1)
        self['nPolars']=1

        # --- Detect columns
        nCols = self['data'].shape[1]
        # First, if all values are numeric, abort
        onestring = header.replace(',',' ')
        try:
            vals = np.array(onestring.split()).astype(float) # This should fail
        except:
            pass # Great, it actually failed, the first line is not made of floats
        else:
            raise WrongPolarFormatError('The first line is all numeric, it should contain column names')
        # Then, try to split by commas or space
        colsComma = header.split(',')
        colsSpace = header.split()
        if len(colsComma)==nCols:
            cols = [c.strip() for c in colsComma]
        elif len(colsSpace)==nCols:
            cols = colsSpace
        else:
            raise BrokenPolarFormatError('The number of header columns ({}) does not match the number of columns in the data ({})'.format(len(cols),nCols))
        self['columns'] = cols

class PolarFile_NoHeader(BasePolarFile):
    """ 
    Polar file with no header, or some "meaningless" comments that starts with ['#','!','%']
    Data may be space or column separated
    """
    @staticmethod
    def formatName(): return 'Polar file no header'

    def read(self, filename):
        self['data']    = _load_txt(filename, self.COMMENT_CHARS)
        self['nPolars'] = 1
        # --- Detect columns
        nCols = self['data'].shape[1]
        d = [DEFAULT_COLUMNS['alpha'], DEFAULT_COLUMNS['cl'], DEFAULT_COLUMNS['cd'], DEFAULT_COLUMNS['cm']]
        n2col = {2:d[0:2], 3:d[0:3], 4:d[0:4] }
        if nCols in n2col.keys():
            self['columns'] = n2col[nCols]
        else:
            raise BrokenPolarFormatError('The number of columns in the data ({}) is not amongst the supported ones ({}).'.format(nCols, n2col.keys()))

class PolarFile_AD_Basic(BasePolarFile):
    """ 
    Reads a basic AeroDyn file
    """
    @staticmethod
    def formatName(): return 'Polar AeroDyn file basic'

    def read(self, filename):
        self['data']    = _load_txt(filename, self.COMMENT_CHARS, skiprows = 53)
        self['nPolars'] = 1
        # import pandas as pd
        # df=pd.read_csv(filename, skiprows = 53, header=None, delim_whitespace=True, names=['Alpha','Cl','Cd','Cm']).values
        # --- Detect columns
        nCols = self['data'].shape[1]
        n2col = {2:['Alpha','Cl'], 3:['Alpha','Cl', 'Cm'], 4:['Alpha','Cl', 'Cm', 'Cd'] }
        if nCols in n2col.keys():
            self['columns'] = n2col[nCols]
        else:
            raise BrokenPolarFormatError('The number of columns in the data ({}) is not amongst the supported ones ({}).'.format(nCols, n2col.keys()))


class PolarFile(BasePolarFile):
    """ """
    @staticmethod
    def formatName(): return 'Polar file'


def loadPolarFile(filename, fformat='auto', to_radians=False, standardizeCols=True, verbose=False):
    """ 
    Loads a PolarFile, return a dataFrame
    """
    if not os.path.exists(filename):
        raise Exception('File not found:',filename)
        print('[WARN] Not all file formats supported ')

    allReaders   = [ADPolarFile, PolarFile_OneLineHeader, PolarFile_NoHeader, PolarFile_AD_Basic, CSVFile]
    delimReaders = [PolarFile_OneLineHeader, PolarFile_AD_Basic, CSVFile]

    def tryReading(f, reader):
        if f is not None:
            return f
        if reader is None:
            return None
        try:
            if verbose:
                print('')
                print('PolarFile: trying to read with format: {}'.format(reader.formatName()))
            return reader(filename)
        except:
            if verbose:
                print('>>> PolarFile: Failed to read with format: {}'.format(reader.formatName()))
            pass
    f = None
    Re = np.nan # TODO

    if fformat==None:
        fformat = 'auto'

    if fformat=='ADPolar':
        f = ADPolarFile(filename)

    elif fformat=='delimited':

        for reader in delimReaders:
            f = tryReading(f, reader)
            if f is not None:
                break

    elif fformat=='auto':

        for reader in allReaders:
            f = tryReading(f, reader)
            if f is not None:
                break

    if f is None:
        raise Exception('Unable to read the polar {} using the fileformat {}. Use a supported fileformat'.format(filename, fformat))

    # --- Store in DataFrame
    df = f.toDataFrame()
    if verbose:
        print('PolarFile: Columns before: ',df.columns.values)

    # --- Rename columns - Standardize column names
    if standardizeCols:
        COLS_TODO= {**DEFAULT_COLUMNS,**DEFAULT_COLUMNS_EXT}
        for ic, col in enumerate(df.columns):
            c = col.strip().lower().replace('_','')
            c = c.replace('aoa','alpha')
            c = c.replace('fst','fs')
            c = c.replace('012','z12')
            c = c.replace('cllin','clinv')
            c = c.replace('clpot','clinv')
            known_keys = reversed(sorted(list(COLS_TODO.keys())))
            found=False
            for kk in known_keys:
                if c.startswith(kk):
                    cnew = COLS_TODO.pop(kk)
                    #print('changing {} to {}'.format(c, cnew))
                    df.columns.values[ic] = cnew # rename column
                    found=True
                    break
            if not found:
                print('[WARN] PolarFile: The following column was not understood: {}'.format(col))

        # --- Standardize data
        for k,v in DEFAULT_COLUMNS.items():
            if v not in df.columns:
                 df[v] = np.nan
    if verbose:
        print('PolarFile: Columns after: ',df.columns.values)

    if standardizeCols:
        cAlpha = DEFAULT_COLUMNS['alpha']
        if cAlpha not in df.columns:
            raise Exception('Angle of attack was not detected as part of the columns')
    else:
        cAlpha = df.columns.values[0]

    if to_radians:
        # First, check the data, if the max alpha is above pi, most likely we are in degrees
        _radians = np.mean(np.abs(df[cAlpha])) <= np.pi / 2
        if _radians:
            raise Exception('PolarFile: Asked to convert input to radian, but the data is likely already in radians.')
        df[cAlpha]*=np.pi/180

    Re  = np.nan
    return df, Re 

if __name__ == "__main__":
    from welib.tools.clean_exceptions import *
#     PolarFile_OneLineHeader('data/63-235.csv')
    #f = PolarFile_NoHeader('data/63-235.csv')
#     f = loadPolarFile('data/63-235.csv')
    f = loadPolarFile('data/FFA-W3-241-Re12M.dat', verbose=True)
    #f = loadPolarFile('data/Cylinder.dat')
    #f = loadPolarFile('../../data/NREL5MW/5MW_Baseline/Airfoils/DU21_A17.dat')
    print(f)
    pass


import unittest

import numpy as np
import numpy.testing as npt

import wisdem.commonse.utilization_api as util
from wisdem.commonse import gravity as g

myones = np.ones((100,))


class TestUtilization(unittest.TestCase):
    def testTBeam(self):
        h_web = 10.0
        w_flange = 8.0
        t_web = 3.0
        t_flange = 4.0

        area, y_cg, Ixx, Iyy = util._TBeamProperties(h_web, t_web, w_flange, t_flange)
        self.assertEqual(area, 62.0)
        self.assertAlmostEqual(y_cg, 8.6129, 4)
        self.assertAlmostEqual(Iyy, 193.16666, 4)
        self.assertAlmostEqual(Ixx, 1051.37631867699, 4)

        area, y_cg, Ixx, Iyy = util._TBeamProperties(
            h_web * myones, t_web * myones, w_flange * myones, t_flange * myones
        )
        npt.assert_equal(area, 62.0 * myones)
        npt.assert_almost_equal(y_cg, 8.6129 * myones, 1e-4)
        npt.assert_almost_equal(Iyy, 193.16666 * myones, 1e-4)
        npt.assert_almost_equal(Ixx, 1051.37631867699 * myones, 1e-4)

    def testPlasticityRF(self):
        Fy = 4.0

        Fe = 1.0
        Fi = Fe
        self.assertEqual(util._plasticityRF(Fe, Fy), Fi)
        npt.assert_equal(util._plasticityRF(Fe * myones, Fy), Fi * myones)

        Fe = 3.0
        Fr = 4.0 / 3.0
        Fi = Fe * Fr * (1.0 + 3.75 * Fr**2) ** (-0.25)
        self.assertEqual(util._plasticityRF(Fe, Fy), Fi)
        npt.assert_equal(util._plasticityRF(Fe * myones, Fy), Fi * myones)

    def testSafetyFactor(self):
        Fy = 100.0
        k = 1.25
        self.assertEqual(util._safety_factor(25.0, Fy), k * 1.2)
        npt.assert_equal(util._safety_factor(25.0 * myones, Fy), k * 1.2 * myones)
        self.assertEqual(util._safety_factor(125.0, Fy), k * 1.0)
        npt.assert_equal(util._safety_factor(125.0 * myones, Fy), k * 1.0 * myones)
        self.assertAlmostEqual(util._safety_factor(80.0, Fy), k * 1.08)
        npt.assert_almost_equal(util._safety_factor(80.0 * myones, Fy), k * 1.08 * myones)

    def testAppliedHoop(self):
        # Use the API 2U Appendix B as a big unit test!
        ksi_to_si = 6894757.29317831
        lbperft3_to_si = 16.0185
        ft_to_si = 0.3048
        in_to_si = ft_to_si / 12.0

        R_od = 0.5 * 600 * in_to_si
        t_wall = 0.75 * in_to_si
        rho = 64.0 * lbperft3_to_si
        z = 60 * ft_to_si
        pressure = rho * g * z
        expect = 1e-3 * 64.0 * 60.0 / 144.0 * ksi_to_si

        self.assertAlmostEqual(pressure, expect, -4)
        expect *= R_od / t_wall
        self.assertAlmostEqual(util._compute_applied_hoop(pressure, R_od, t_wall), expect, -4)
        npt.assert_almost_equal(
            util._compute_applied_hoop(pressure * myones, R_od * myones, t_wall * myones), expect * myones, decimal=-4
        )

    """
    def testAppliedAxial(self):
        # Use the API 2U Appendix B as a big unit test!
        ksi_to_si = 6894757.29317831
        lbperft3_to_si = 16.0185
        ft_to_si = 0.3048
        in_to_si = ft_to_si / 12.0
        kip_to_si = 4.4482216 * 1e3

        R_od = 0.5*600 * np.ones((4,)) * in_to_si
        t_wall = 0.75 * np.ones((4,)) * in_to_si
        t_web = 5./8. * np.ones((3,)) * in_to_si
        h_web = 14.0 * np.ones((3,)) * in_to_si
        t_flange = 1.0 * np.ones((3,)) * in_to_si
        w_flange = 10.0 * np.ones((3,)) * in_to_si
        h_section = 50.0 * np.ones((3,)) * ft_to_si
        L_stiffener = 5.0 * np.ones((3,)) * ft_to_si
        self.params['water_density'] = 64.0 * lbperft3_to_si
        E = 29e3 * ksi_to_si
        nu = 0.3
        sigma_y = 50 * ksi_to_si
        self.params['bulkhead_nodes'] = [False, False, False, False]
        self.params['wave_height'] = 0.0 # gives only static pressure
        self.params['stack_mass_in'] = 9000 * kip_to_si/g

        self.set_geometry()
        self.myspar.section_mass = np.zeros((3,))

        expect = 9000 * kip_to_si / (2*np.pi*(0.5*R_od[0]-0.5*t_wall[0])*t_wall[0])
        npt.assert_almost_equal(util._compute_applied_axial(self.params, self.myspar.section_mass), expect* np.ones((3,)), decimal=4)
    """

    def testStiffenerFactors(self):
        # Use the API 2U Appendix B as a big unit test!
        ksi_to_si = 6894757.29317831
        lbperft3_to_si = 16.0185
        ft_to_si = 0.3048
        in_to_si = ft_to_si / 12.0
        kip_to_si = 4.4482216 * 1e3

        R_od = 0.5 * 600 * np.ones((3,)) * in_to_si
        t_wall = 0.75 * np.ones((3,)) * in_to_si
        t_web = 5.0 / 8.0 * np.ones((3,)) * in_to_si
        h_web = 14.0 * np.ones((3,)) * in_to_si
        t_flange = 1.0 * np.ones((3,)) * in_to_si
        w_flange = 10.0 * np.ones((3,)) * in_to_si
        L_stiffener = 5.0 * np.ones((3,)) * ft_to_si
        E = 29e3 * ksi_to_si * np.ones((3,))
        nu = 0.3 * np.ones((3,))

        pressure = 1e-3 * 64.0 * 60.0 / 144.0 * ksi_to_si
        axial = 9000 * kip_to_si / (2 * np.pi * (R_od[0] - 0.5 * t_wall[0]) * t_wall[0])
        self.assertAlmostEqual(axial, 0.5 * 9000 / 299.625 / 0.75 / np.pi * ksi_to_si, -4)
        KthL, KthG = util._compute_stiffener_factors(
            pressure, axial, R_od, t_wall, h_web, t_web, w_flange, t_flange, L_stiffener, E, nu
        )
        npt.assert_almost_equal(KthL, 1.0 * np.ones((3,)), decimal=1)
        npt.assert_almost_equal(KthG, 0.5748 * np.ones((3,)), decimal=4)  # 0.5642 if R_flange accounts for t_wall

    def testStressLimits(self):
        # Use the API 2U Appendix B as a big unit test!
        ksi_to_si = 6894757.29317831
        lbperft3_to_si = 16.0185
        ft_to_si = 0.3048
        in_to_si = ft_to_si / 12.0
        kip_to_si = 4.4482216 * 1e3

        R_od = 0.5 * 600 * np.ones((3,)) * in_to_si
        t_wall = 0.75 * np.ones((3,)) * in_to_si
        t_web = 5.0 / 8.0 * np.ones((3,)) * in_to_si
        h_web = 14.0 * np.ones((3,)) * in_to_si
        t_flange = 1.0 * np.ones((3,)) * in_to_si
        w_flange = 10.0 * np.ones((3,)) * in_to_si
        L_stiffener = 5.0 * np.ones((3,)) * ft_to_si
        h_section = 50.0 * np.ones((3,)) * ft_to_si
        E = 29e3 * ksi_to_si * np.ones((3,))
        nu = 0.3 * np.ones((3,))
        sigma_y = 50 * ksi_to_si * np.ones((3,))

        KthG = 0.5748
        FxeL, FreL, FxeG, FreG = util._compute_elastic_stress_limits(
            R_od, t_wall, h_section, h_web, t_web, w_flange, t_flange, L_stiffener, E, nu, KthG, loading="radial"
        )
        npt.assert_almost_equal(FxeL, 16.074844135928885 * ksi_to_si * np.ones((3,)), decimal=1)
        npt.assert_almost_equal(FreL, 19.80252150945599 * ksi_to_si * np.ones((3,)), decimal=1)
        npt.assert_almost_equal(FxeG, 37.635953475479639 * ksi_to_si * np.ones((3,)), decimal=1)
        npt.assert_almost_equal(FreG, 93.77314503852581 * ksi_to_si * np.ones((3,)), decimal=1)

        FxcL = util._plasticityRF(FxeL, sigma_y)
        FxcG = util._plasticityRF(FxeG, sigma_y)
        FrcL = util._plasticityRF(FreL, sigma_y)
        FrcG = util._plasticityRF(FreG, sigma_y)
        npt.assert_almost_equal(FxcL, 1.0 * 16.074844135928885 * ksi_to_si * np.ones((3,)), decimal=1)
        npt.assert_almost_equal(FrcL, 1.0 * 19.80252150945599 * ksi_to_si * np.ones((3,)), decimal=1)
        npt.assert_almost_equal(FxcG, 0.799647237534 * 37.635953475479639 * ksi_to_si * np.ones((3,)), decimal=1)
        npt.assert_almost_equal(FrcG, 0.444735273606 * 93.77314503852581 * ksi_to_si * np.ones((3,)), decimal=1)

    def testCheckStresses(self):
        # Use the API 2U Appendix B as a big unit test!
        ksi_to_si = 6894757.29317831
        lbperft3_to_si = 16.0185
        ft_to_si = 0.3048
        in_to_si = ft_to_si / 12.0
        kip_to_si = 4.4482216 * 1e3

        R_od = 0.5 * 600 * np.ones((3,)) * in_to_si
        t_wall = 0.75 * np.ones((3,)) * in_to_si
        t_web = 5.0 / 8.0 * np.ones((3,)) * in_to_si
        h_web = 14.0 * np.ones((3,)) * in_to_si
        t_flange = 1.0 * np.ones((3,)) * in_to_si
        w_flange = 10.0 * np.ones((3,)) * in_to_si
        h_section = 50.0 * np.ones((3,)) * ft_to_si
        L_stiffener = 5.0 * np.ones((3,)) * ft_to_si
        rho = 64.0 * lbperft3_to_si
        E = 29e3 * ksi_to_si * np.ones((3,))
        nu = 0.3 * np.ones((3,))
        sigma_y = 50 * ksi_to_si * np.ones((3,))

        # Find pressure to give "head" of 60ft- put mid-point of middle section at this depth
        z = 60 * ft_to_si
        P = rho * g * z
        sigma_ax = 9000 * kip_to_si / (2 * np.pi * (R_od[0] - 0.5 * t_wall[0]) * t_wall[0])

        (
            axial_local_unity,
            axial_general_unity,
            external_local_unity,
            external_general_unity,
            _,
            _,
            _,
            _,
        ) = util.shellBuckling_withStiffeners(
            P,
            sigma_ax,
            R_od,
            t_wall,
            h_section,
            h_web,
            t_web,
            w_flange,
            t_flange,
            L_stiffener,
            E,
            nu,
            sigma_y,
            loading="radial",
        )

        # npt.assert_almost_equal(web_compactness, 24.1/22.4 * np.ones((3,)), decimal=3)
        # npt.assert_almost_equal(flange_compactness, 9.03/5.0 * np.ones((3,)), decimal=3)
        npt.assert_almost_equal(axial_local_unity, 1.07, 1)
        npt.assert_almost_equal(axial_general_unity, 0.34, 1)
        npt.assert_almost_equal(external_local_unity, 1.07, 1)
        npt.assert_almost_equal(external_general_unity, 0.59, 1)


if __name__ == "__main__":
    unittest.main()



"""`ArrayCableInstallation` class and related processes."""

__author__ = "Jake Nunemaker"
__copyright__ = "Copyright 2020, National Renewable Energy Laboratory"
__maintainer__ = "Jake Nunemaker"
__email__ = "jake.nunemaker@nrel.gov"


from copy import deepcopy

import numpy as np
from marmot import process

from wisdem.orbit.core import Vessel
from wisdem.orbit.core.logic import position_onsite
from wisdem.orbit.phases.install import InstallPhase
from wisdem.orbit.core.exceptions import InsufficientCable

from .common import SimpleCable as Cable
from .common import (
    lay_cable,
    bury_cable,
    dig_trench,
    prep_cable,
    lower_cable,
    pull_in_cable,
    lay_bury_cable,
    terminate_cable,
    load_cable_on_vessel,
)


class ArrayCableInstallation(InstallPhase):
    """Array Cable Installation Phase"""

    phase = "Array Cable Installation"
    capex_category = "Array System"

    #:
    expected_config = {
        "array_cable_install_vessel": "str",
        "array_cable_bury_vessel": "str (optional)",
        "array_cable_trench_vessel": "str (optional)",
        "site": {"distance": "km", "depth": "m"},
        "array_system": {
            "system_cost": "USD",
            "num_strings": "int (optional, default: 10)",
            "free_cable_length": "km (optional, default: 'depth')",
            "cables": {
                "name (variable)": {
                    "linear_density": "t/km",
                    "cable_sections": [("length, km", "int", "speed, km/h (optional)")],
                }
            },
        },
    }

    def __init__(self, config, weather=None, **kwargs):
        """
        Creates an instance of ArrayCableInstallation.

        Parameters
        ----------
        config : dict
            Simulation specific configuration.
        weather : np.ndarray
            Weather profile at site.
        """

        super().__init__(weather, **kwargs)

        config = self.initialize_library(config, **kwargs)
        self.config = self.validate_config(config)

        self.initialize_port()
        self.setup_simulation(**kwargs)

    def setup_simulation(self, **kwargs):
        """
        Setup method for ArrayCableInstallation phase.
        - Extracts key inputs
        -
        """

        depth = self.config["site"]["depth"]
        system = self.config["array_system"]
        self.free_cable_length = system.get("free_cable_length", depth / 1000)

        self.initialize_installation_vessel()
        self.initialize_burial_vessel()
        self.initialize_trench_vessel()

        self.num_strings = system.get("num_strings", 10)
        self.cable_data = [
            (Cable(data["linear_density"]), deepcopy(data["cable_sections"])) for _, data in system["cables"].items()
        ]

        # Perform cable installation
        install_array_cables(
            self.install_vessel,
            distance=self.config["site"]["distance"],
            cable_data=self.cable_data,
            num_strings=self.num_strings,
            burial_vessel=self.bury_vessel,
            trench_vessel=self.trench_vessel,
            free_cable_length=self.free_cable_length,
            **kwargs,
        )

    @property
    def system_capex(self):
        """Returns total procurement cost of the array system."""

        return self.config["array_system"]["system_cost"]

    def initialize_installation_vessel(self):
        """Creates the array cable installation vessel."""

        # Vessel name and costs
        vessel_specs = self.config.get("array_cable_install_vessel", None)
        name = vessel_specs.get("name", "Array Cable Installation Vessel")

        vessel = self.initialize_vessel(name, vessel_specs)
        self.env.register(vessel)

        vessel.initialize()
        vessel.at_port = True
        vessel.at_site = False
        self.install_vessel = vessel

    def initialize_burial_vessel(self):
        """Creates the array cable burial vessel."""

        # Vessel name and costs
        vessel_specs = self.config.get("array_cable_bury_vessel", None)
        if vessel_specs is None:
            self.bury_vessel = None
            return
        name = vessel_specs.get("name", "Array Cable Burial Vessel")

        vessel = self.initialize_vessel(name, vessel_specs)
        self.env.register(vessel)

        vessel.initialize()
        vessel.at_port = True
        vessel.at_site = False
        self.bury_vessel = vessel

    def initialize_trench_vessel(self):
        """Creates the array cable trenching vessel."""

        # Vessel name and costs
        vessel_specs = self.config.get("array_cable_trench_vessel", None)
        if vessel_specs is None:
            self.trench_vessel = None
            return
        name = vessel_specs.get("name", "Array Cable Trench Vessel")

        vessel = self.initialize_vessel(name, vessel_specs)
        self.env.register(vessel)

        vessel.initialize()
        vessel.at_port = True
        vessel.at_site = False
        self.trench_vessel = vessel

    @property
    def detailed_output(self):
        """Detailed outputs of the array system installation."""

        outputs = {self.phase: {**self.agent_efficiencies}}

        return outputs


@process
def install_array_cables(
    vessel,
    distance,
    cable_data,
    num_strings,
    burial_vessel=None,
    trench_vessel=None,
    free_cable_length=None,
    **kwargs,
):
    """
    Simulation of the installation of array cables.

    Parameters
    ----------
    vessel : Vessel
        Cable installation vessel.
    cable_data : list
        List of tuples containing `Cable` instances and sections.
    num_strings : int
        Number of array system strings. Used for partial generation assumptions
        for the post-processed cash flow model.
    burial_vessel : Vessel
        Optional configuration for burial vessel. If configured, the
        installation vessel only lays the cable on the seafloor and this
        vessel will bury them at the end of the simulation.
    trench_vessel: Vessel
        Optional configuration for trenching vessel.  If configured, the
        trenching vessel travels along the cable route prior to arrival of
        the cable lay vessel and digs a trench.
    """

    breakpoints = list(np.linspace(1 / num_strings, 1, num_strings))
    trench_sections = []
    total_cable_length = 0
    installed = 0

    for cable, sections in cable_data:
        for s in sections:
            l, num_i, *_ = s
            total_cable_length += l * num_i

            _trench_length = max(0, l - 2 * free_cable_length)
            if _trench_length:
                trench_sections.extend([_trench_length] * num_i)

    ## Trenching Process
    # Conduct trenching along cable routes before laying cable
    if trench_vessel is None:
        pass

    else:
        # Conduct trenching operations
        while True:
            if trench_vessel.at_port:
                trench_vessel.at_port = False
                yield trench_vessel.transit(distance, **kwargs)
                trench_vessel.at_site = True

            elif trench_vessel.at_site:
                try:
                    # Dig trench along each cable section distance
                    trench_distance = trench_sections.pop(0)
                    yield dig_array_cables_trench(trench_vessel, trench_distance, **kwargs)

                except IndexError:
                    trench_vessel.at_site = False
                    yield trench_vessel.transit(distance, **kwargs)
                    trench_vessel.at_port = True
                    break

        vessel.submit_debug_log(message="Array cable trench digging process completed!")

    ## Cable Lay Process
    to_bury = []
    for cable, sections in cable_data:
        vessel.cable_storage.reset()

        while True:
            if vessel.at_port:
                yield load_cable_on_vessel(vessel, cable, **kwargs)

                vessel.at_port = False
                yield vessel.transit(distance, **kwargs)
                vessel.at_site = True

            elif vessel.at_site:
                try:
                    length, num_sections, *extra = sections.pop(0)
                    if extra:
                        speed = extra[0]

                        if burial_vessel is None:
                            specs = {**kwargs, "cable_lay_bury_speed": speed}

                        else:
                            specs = {**kwargs, "cable_lay_speed": speed}

                    else:
                        specs = deepcopy(kwargs)

                except IndexError:
                    vessel.at_site = False
                    yield vessel.transit(distance, **kwargs)
                    vessel.at_port = True
                    break

                for _ in range(num_sections):
                    try:
                        section = vessel.cable_storage.get_cable(length)

                    except InsufficientCable:
                        yield vessel.transit(distance, **kwargs)
                        yield load_cable_on_vessel(vessel, cable, **kwargs)
                        yield vessel.transit(distance, **kwargs)
                        section = vessel.cable_storage.get_cable(length)

                    # Prep for cable laying procedure (at substructure 1)
                    yield position_onsite(vessel, **kwargs)
                    yield prep_cable(vessel, **kwargs)
                    yield pull_in_cable(vessel, **kwargs)
                    yield terminate_cable(vessel, **kwargs)
                    yield lower_cable(vessel, **kwargs)

                    # Cable laying procedure
                    if burial_vessel is None:
                        yield lay_bury_cable(vessel, section, **specs)
                        installed += section

                    else:
                        yield lay_cable(vessel, section, **specs)
                        _bury = max(0, (section - 2 * free_cable_length))
                        if _bury:
                            to_bury.append(_bury)

                    # Post cable laying procedure (at substructure 2)
                    yield prep_cable(vessel, **kwargs)
                    yield pull_in_cable(vessel, **kwargs)
                    yield terminate_cable(vessel, **kwargs)

                    if burial_vessel is None:
                        breakpoints = check_for_completed_string(vessel, installed, total_cable_length, breakpoints)

        # Transit back to port
        vessel.at_site = False
        yield vessel.transit(distance, **kwargs)
        vessel.at_port = True

    ## Burial Process
    if burial_vessel is None:
        vessel.submit_debug_log(message="Array cable lay/burial process completed!")

    else:
        vessel.submit_debug_log(message="Array cable lay process completed!")
        bury_array_cables(burial_vessel, to_bury, breakpoints, **kwargs)


@process
def bury_array_cables(vessel, sections, breakpoints, **kwargs):
    """
    Simulation for the burial of array cables if configured.

    Parameters
    ----------
    vessel : Vessel
        Performing vessel.
    sections : list
        List of cable sections that need to be buried at site.
    breakpoints : list
        TODO
        List of string breakpoints.
    """

    installed = 0
    total_length = sum(sections)

    for length in sections:
        yield position_onsite(vessel, site_position_time=2)
        yield bury_cable(vessel, length, **kwargs)
        installed += length

        breakpoints = check_for_completed_string(vessel, installed, total_length, breakpoints)

    vessel.submit_debug_log(message="Array cable burial process completed!")


@process
def dig_array_cables_trench(vessel, distance, **kwargs):
    """
    Simulation for digging a trench for the array cables (if configured).

    Parameters
    ----------
    vessel : Vessel
        Performing vessel.
    distance : int | float
        Distance between turbines to dig trench for array cable
    """

    yield position_onsite(vessel, site_position_time=2)
    yield dig_trench(vessel, distance, **kwargs)


def check_for_completed_string(vessel, installed, total, breakpoints):
    """
    TODO:
    """

    if (installed / total) >= breakpoints[0]:
        vessel.submit_debug_log(progress="Array String")
        _ = breakpoints.pop(0)

    return breakpoints



#!/usr/bin/env python
# encoding: utf-8

import numpy as np
import openmdao.api as om

import wisdem.pyframe3dd.pyframe3dd as frame3dd
from wisdem.commonse import gravity
from wisdem.commonse.csystem import DirectionVector
from wisdem.commonse.utilities import find_nearest, nodal2sectional
from wisdem.commonse.cross_sections import Tube, IBeam
from wisdem.commonse.utilization_constraints import TubevonMisesStressUtilization

RIGID = 1
FREE = 0


def tube_prop(s, Di, ti):
    L = s.max() - s.min()

    def equal_pts(xi):
        if len(xi) < len(s) and len(xi) == 2:
            x = np.interp((s - s.min()) / L, [0, 1], xi)
        elif len(xi) == len(s):
            x = xi
        else:
            raise ValueError("Unknown grid of input", str(xi))
        return x

    D = equal_pts(Di)
    t = equal_pts(ti)
    return Tube(nodal2sectional(D)[0], nodal2sectional(t)[0])


class Hub_Rotor_LSS_Frame(om.ExplicitComponent):
    """
    Run structural analysis of hub system with the generator rotor and main (LSS) shaft.

    Parameters
    ----------
    tilt : float, [deg]
        Shaft tilt
    s_lss : numpy array[5], [m]
        Discretized s-coordinates along drivetrain, measured from bedplate (direct) or tower center (geared)
    lss_diameter : numpy array[2], [m]
        LSS outer diameter from hub to bearing 2
    lss_wall_thickness : numpy array[2], [m]
        LSS wall thickness
    hub_system_mass : float, [kg]
        Hub system mass
    hub_system_cm : float, [m]
        Hub system center of mass distance from hub flange
    hub_system_I : numpy array[6], [kg*m**2]
        Hub system moment of inertia
    F_hub : numpy array[3, n_dlcs], [N]
        Force vector applied to the hub (WITH WEIGHT???)
    M_hub : numpy array[3, n_dlcs], [N*m]
        Moment vector applied to the hub
    s_mb1 : float, [m]
        Bearing 1 s-coordinate along drivetrain, measured from bedplate (direct) or tower center (geared)
    s_mb2 : float, [m]
        Bearing 2 s-coordinate along drivetrain, measured from bedplate (direct) or tower center (geared)
    s_rotor : float, [m]
        Generator rotor attachment to lss s-coordinate measured from bedplate (direct) or tower center (geared)
    generator_rotor_mass : float, [kg]
        Generator rotor mass
    generator_rotor_I : numpy array[3], [kg*m**2]
        Generator rotor moment of inertia (measured about its cm)
    gearbox_mass : float, [kg]
        Gearbox rotor mass
    gearbox_I : numpy array[3], [kg*m**2]
        Gearbox moment of inertia (measured about its cm)
    lss_E : float, [Pa]
        modulus of elasticity
    lss_G : float, [Pa]
        shear modulus
    lss_rho : float, [kg/m**3]
        material density
    lss_Xy : float, [Pa]
        yield stress

    Returns
    -------
    lss_spring_constant : float, [N*m/rad]
        Equivalent spring constant for the low speed shaft froom T=(G*J/L)*theta
    torq_deflection : float, [m]
        Maximum deflection distance at rotor (direct) or gearbox (geared) attachment
    torq_angle : float, [rad]
        Maximum rotation angle at rotor (direct) or gearbox (geared) attachment
    torq_axial_stress : numpy array[5, n_dlcs], [Pa]
        Axial stress in Curved_beam structure
    torq_shear_stress : numpy array[5, n_dlcs], [Pa]
        Shear stress in Curved_beam structure
    torq_bending_stress : numpy array[5, n_dlcs], [Pa]
        Hoop stress in Curved_beam structure calculated with Roarks formulae
    constr_lss_vonmises : numpy array[5, n_dlcs]
        Sigma_y/Von_Mises
    F_mb1 : numpy array[3, n_dlcs], [N]
        Force vector applied to bearing 1 in hub c.s.
    F_mb2 : numpy array[3, n_dlcs], [N]
        Force vector applied to bearing 2 in hub c.s.
    F_torq : numpy array[3, n_dlcs], [N]
        Force vector applied to generator rotor (direct) or gearbox (geared) in hub c.s.
    M_mb1 : numpy array[3, n_dlcs], [N*m]
        Moment vector applied to bearing 1 in hub c.s.
    M_mb2 : numpy array[3, n_dlcs], [N*m]
        Moment vector applied to bearing 2 in hub c.s.
    M_torq : numpy array[3, n_dlcs], [N*m]
        Moment vector applied to generator rotor (direct) or gearbox (geared) in hub c.s.
    lss_axial_load2stress : numpy array[nFull-1,6], [m**2]
        Linear conversion factors between loads [Fx-z; Mx-z] and axial stress
    lss_shear_load2stress : numpy array[nFull-1,6], [m**2]
        Linear conversion factors between loads [Fx-z; Mx-z] and shear stress

    """

    def initialize(self):
        self.options.declare("n_dlcs")
        self.options.declare("direct_drive", default=True)
        self.options.declare("modeling_options")

    def setup(self):
        n_dlcs = self.options["n_dlcs"]

        self.add_discrete_input("upwind", True)
        self.add_input("tilt", 0.0, units="deg")
        self.add_input("s_lss", val=np.zeros(5), units="m")
        self.add_input("lss_diameter", val=np.zeros(2), units="m")
        self.add_input("lss_wall_thickness", val=np.zeros(2), units="m")
        self.add_input("hub_system_mass", 0.0, units="kg")
        self.add_input("hub_system_cm", 0.0, units="m")
        self.add_input("hub_system_I", np.zeros(6), units="kg*m**2")
        self.add_input("F_hub", val=np.zeros((3, n_dlcs)), units="N")
        self.add_input("M_hub", val=np.zeros((3, n_dlcs)), units="N*m")
        self.add_input("s_mb1", val=0.0, units="m")
        self.add_input("s_mb2", val=0.0, units="m")
        self.add_input("s_rotor", val=0.0, units="m")
        self.add_input("generator_rotor_mass", val=0.0, units="kg")
        self.add_input("generator_rotor_I", val=np.zeros(3), units="kg*m**2")
        self.add_input("gearbox_mass", val=0.0, units="kg")
        self.add_input("gearbox_I", val=np.zeros(3), units="kg*m**2")
        self.add_input("brake_mass", val=0.0, units="kg")
        self.add_input("brake_I", val=np.zeros(3), units="kg*m**2")
        self.add_input("carrier_mass", val=0.0, units="kg")
        self.add_input("carrier_I", val=np.zeros(3), units="kg*m**2")
        self.add_input("lss_E", val=0.0, units="Pa")
        self.add_input("lss_G", val=0.0, units="Pa")
        self.add_input("lss_rho", val=0.0, units="kg/m**3")
        self.add_input("lss_Xy", val=0.0, units="Pa")

        self.add_input("shaft_deflection_allowable", val=1.0, units="m")
        self.add_input("shaft_angle_allowable", val=1.0, units="rad")

        self.add_output("lss_spring_constant", 0.0, units="N*m/rad")
        self.add_output("torq_deflection", val=0.0, units="m")
        self.add_output("torq_angle", val=0.0, units="rad")
        self.add_output("lss_axial_stress", np.zeros((4, n_dlcs)), units="Pa")
        self.add_output("lss_shear_stress", np.zeros((4, n_dlcs)), units="Pa")
        self.add_output("constr_lss_vonmises", np.zeros((4, n_dlcs)))
        self.add_output("F_mb1", val=np.zeros((3, n_dlcs)), units="N")
        self.add_output("F_mb2", val=np.zeros((3, n_dlcs)), units="N")
        self.add_output("F_torq", val=np.zeros((3, n_dlcs)), units="N")
        self.add_output("M_mb1", val=np.zeros((3, n_dlcs)), units="N*m")
        self.add_output("M_mb2", val=np.zeros((3, n_dlcs)), units="N*m")
        self.add_output("M_torq", val=np.zeros((3, n_dlcs)), units="N*m")
        self.add_output("lss_axial_load2stress", val=np.zeros(6), units="m**2")
        self.add_output("lss_shear_load2stress", val=np.zeros(6), units="m**2")
        self.add_output("constr_shaft_deflection", 0.0)
        self.add_output("constr_shaft_angle", 0.0)

    def compute(self, inputs, outputs, discrete_inputs, discrete_outputs):
        # Unpack inputs
        upwind = discrete_inputs["upwind"]
        Cup = -1.0 if upwind else 1.0
        tilt = float(np.deg2rad(inputs["tilt"][0]))

        s_lss = inputs["s_lss"]
        D_lss = inputs["lss_diameter"]
        t_lss = inputs["lss_wall_thickness"]

        s_mb1 = float(inputs["s_mb1"][0])
        s_mb2 = float(inputs["s_mb2"][0])

        if self.options["direct_drive"]:
            s_rotor = float(inputs["s_rotor"][0])
            m_rotor = float(inputs["generator_rotor_mass"][0])
            I_rotor = inputs["generator_rotor_I"]

            m_brake = float(inputs["brake_mass"][0])
            I_brake = inputs["brake_I"]
        else:
            m_gearbox = float(inputs["gearbox_mass"][0])
            I_gearbox = inputs["gearbox_I"]

            m_carrier = float(inputs["carrier_mass"][0])
            I_carrier = inputs["carrier_I"]

        rho = float(inputs["lss_rho"][0])
        E = float(inputs["lss_E"][0])
        G = float(inputs["lss_G"][0])
        sigma_y = float(inputs["lss_Xy"][0])
        gamma_f = float(self.options["modeling_options"]["gamma_f"])
        gamma_m = float(self.options["modeling_options"]["gamma_m"])
        gamma_n = float(self.options["modeling_options"]["gamma_n"])
        gamma = gamma_f * gamma_m * gamma_n

        m_hub = float(inputs["hub_system_mass"][0])
        cm_hub = float(inputs["hub_system_cm"][0])
        I_hub = inputs["hub_system_I"]
        F_hub = inputs["F_hub"]
        M_hub = inputs["M_hub"]

        torq_defl_allow = float(inputs["shaft_deflection_allowable"][0])
        torq_angle_allow = float(inputs["shaft_angle_allowable"][0])

        # ------- node data ----------------
        n = len(s_lss)
        inode = np.arange(1, n + 1)
        ynode = znode = rnode = np.zeros(n)
        xnode = Cup * s_lss.copy()
        nodes = frame3dd.NodeData(inode, xnode, ynode, znode, rnode)
        # Grab indices for later
        i1 = inode[find_nearest(xnode, Cup * s_mb1)]
        i2 = inode[find_nearest(xnode, Cup * s_mb2)]
        iadd = inode[1]
        # Differences between direct annd geared
        if self.options["direct_drive"]:
            itorq = inode[find_nearest(xnode, Cup * s_rotor)]
            m_torq = m_rotor
            I_torq = I_rotor

            m_add = m_brake
            I_add = I_brake
        else:
            itorq = inode[0]
            m_torq = m_gearbox - m_carrier
            I_torq = I_gearbox - I_carrier

            m_add = m_carrier
            I_add = I_carrier
        # ------------------------------------

        # ------ reaction data ------------
        # Reactions at main bearings
        rnode = np.r_[i1, i2, itorq]
        Rx = np.array([RIGID, FREE, FREE])  # Upwind bearing restricts translational
        Ry = np.array([RIGID, FREE, FREE])  # Upwind bearing restricts translational
        Rz = np.array([RIGID, FREE, FREE])  # Upwind bearing restricts translational
        Rxx = np.array([FREE, FREE, RIGID])  # Torque is absorbed by stator, so this is the best way to capture that
        Ryy = np.array([FREE, RIGID, FREE])  # downwind bearing carry moments
        Rzz = np.array([FREE, RIGID, FREE])  # downwind bearing carry moments
        reactions = frame3dd.ReactionData(rnode, Rx, Ry, Rz, Rxx, Ryy, Rzz, rigid=RIGID)
        # -----------------------------------

        # ------ frame element data ------------
        lsscyl = tube_prop(s_lss, D_lss, t_lss)
        ielement = np.arange(1, n)
        N1 = np.arange(1, n)
        N2 = np.arange(2, n + 1)
        roll = np.zeros(n - 1)
        myones = np.ones(n - 1)
        Ax = lsscyl.Area
        As = lsscyl.Asx
        S = lsscyl.S
        C = lsscyl.C
        J0 = lsscyl.J0
        Jx = lsscyl.Ixx

        elements = frame3dd.ElementData(
            ielement, N1, N2, Ax, As, As, J0, Jx, Jx, E * myones, G * myones, roll, rho * myones
        )

        lss_stiff = G * J0 / np.diff(s_lss)
        outputs["lss_spring_constant"] = 1.0 / np.sum(1.0 / lss_stiff)
        # -----------------------------------

        # ------ options ------------
        shear = geom = True
        dx = -1
        options = frame3dd.Options(shear, geom, dx)
        # -----------------------------------

        # initialize frameDD3 object
        myframe = frame3dd.Frame(nodes, reactions, elements, options)

        # ------ add hub and generator rotor (direct) or gearbox (geared) extra mass ------------
        three0 = np.zeros(3).tolist()
        myframe.changeExtraNodeMass(
            np.r_[inode[-1], itorq, iadd],
            [m_hub, m_torq, m_add],
            [I_hub[0], I_torq[0], I_add[0]],
            [I_hub[1], I_torq[1], I_add[1]],
            [I_hub[2], I_torq[2], I_add[2]],
            three0,
            three0,
            three0,
            [cm_hub, 0.0, 0.0],
            three0,
            three0,
            True,
        )
        # ------------------------------------

        # ------- NO dynamic analysis ----------
        # myframe.enableDynamics(NFREQ, discrete_inputs['Mmethod'], discrete_inputs['lump'], float(inputs['tol']), float(inputs['shift']))
        # ----------------------------

        # ------ static load cases ------------
        n_dlcs = self.options["n_dlcs"]
        gy = 0.0
        gx = -gravity * np.sin(tilt)
        gz = -gravity * np.cos(tilt)
        for k in range(n_dlcs):
            # gravity in the X, Y, Z, directions (global)
            load = frame3dd.StaticLoadCase(gx, gy, gz)

            # point loads
            # TODO: Are input loads aligned with the lss? If so they need to be rotated.
            load.changePointLoads(
                [inode[-1]], [F_hub[0, k]], [F_hub[1, k]], [F_hub[2, k]], [M_hub[0, k]], [M_hub[1, k]], [M_hub[2, k]]
            )
            # -----------------------------------

            # Put all together and run
            myframe.addLoadCase(load)

        # myframe.write('myframe1.3dd') # Debugging
        displacements, forces, reactions, internalForces, mass3dd, modal = myframe.run()

        # Loop over DLCs and append to outputs
        rotor_gearbox_deflection = np.zeros(n_dlcs)
        rotor_gearbox_angle = np.zeros(n_dlcs)
        outputs["F_mb1"] = np.zeros((3, n_dlcs))
        outputs["F_mb2"] = np.zeros((3, n_dlcs))
        outputs["F_torq"] = np.zeros((3, n_dlcs))
        outputs["M_mb1"] = np.zeros((3, n_dlcs))
        outputs["M_mb2"] = np.zeros((3, n_dlcs))
        outputs["M_torq"] = np.zeros((3, n_dlcs))
        outputs["lss_axial_stress"] = np.zeros((n - 1, n_dlcs))
        outputs["lss_shear_stress"] = np.zeros((n - 1, n_dlcs))
        outputs["constr_lss_vonmises"] = np.zeros((n - 1, n_dlcs))
        for k in range(n_dlcs):
            # Deflections and rotations at torq attachment
            rotor_gearbox_deflection[k] = np.sqrt(
                displacements.dx[k, itorq - 1] ** 2
                + displacements.dy[k, itorq - 1] ** 2
                + displacements.dz[k, itorq - 1] ** 2
            )
            rotor_gearbox_angle[k] = (
                displacements.dxrot[k, itorq - 1]
                + displacements.dyrot[k, itorq - 1]
                + displacements.dzrot[k, itorq - 1]
            )

            # shear and bending, one per element (convert from local to global c.s.)
            Fx = forces.Nx[k, 1::2]
            Vy = forces.Vy[k, 1::2]
            Vz = -forces.Vz[k, 1::2]
            F = np.sqrt(Vz**2 + Vy**2)

            Mxx = forces.Txx[k, 1::2]
            Myy = forces.Myy[k, 1::2]
            Mzz = -forces.Mzz[k, 1::2]
            M = np.sqrt(Myy**2 + Mzz**2)

            # Record total forces and moments
            outputs["F_mb1"][:, k] = -1.0 * np.array([reactions.Fx[k, 0], reactions.Fy[k, 0], reactions.Fz[k, 0]])
            outputs["F_mb2"][:, k] = -1.0 * np.array([reactions.Fx[k, 1], reactions.Fy[k, 1], reactions.Fz[k, 1]])
            outputs["F_torq"][:, k] = -1.0 * np.array([reactions.Fx[k, 2], reactions.Fy[k, 2], reactions.Fz[k, 2]])
            outputs["M_mb1"][:, k] = -1.0 * np.array([reactions.Mxx[k, 0], reactions.Myy[k, 0], reactions.Mzz[k, 0]])
            outputs["M_mb2"][:, k] = -1.0 * np.array([reactions.Mxx[k, 1], reactions.Myy[k, 1], reactions.Mzz[k, 1]])
            outputs["M_torq"][:, k] = -1.0 * np.array([reactions.Mxx[k, 2], reactions.Myy[k, 2], reactions.Mzz[k, 2]])
            outputs["lss_axial_stress"][:, k] = np.abs(Fx) / Ax + M / S
            outputs["lss_shear_stress"][:, k] = 2.0 * F / As + np.abs(Mxx) / C
            hoop = np.zeros(F.shape)

            outputs["constr_lss_vonmises"][:, k] = TubevonMisesStressUtilization(
                outputs["lss_axial_stress"][:, k],
                hoop,
                outputs["lss_shear_stress"][:, k],
                gamma,
                sigma_y,
            )
        outputs["torq_deflection"] = rotor_gearbox_deflection.max()
        outputs["torq_angle"] = rotor_gearbox_angle.max()
        outputs["constr_shaft_deflection"] = gamma * outputs["torq_deflection"] / torq_defl_allow
        outputs["constr_shaft_angle"] = gamma * outputs["torq_angle"] / torq_angle_allow

        # Load->stress conversion for fatigue
        ax_load2stress = np.zeros(6)
        ax_load2stress[0] = 1.0 / Ax[0]
        ax_load2stress[4] = 1.0 / S[0]
        ax_load2stress[5] = 1.0 / S[0]
        sh_load2stress = np.zeros(6)
        sh_load2stress[1] = 1.0 / As[0]
        sh_load2stress[2] = 1.0 / As[0]
        sh_load2stress[3] = 1.0 / C[0]
        outputs["lss_axial_load2stress"] = ax_load2stress
        outputs["lss_shear_load2stress"] = sh_load2stress


class HSS_Frame(om.ExplicitComponent):
    """
    Run structural analysis of high speed shaft (HSS) between gearbox and generator (only for geared configurations).

    Parameters
    ----------
    tilt : float, [deg]
        Shaft tilt
    s_hss : numpy array[3], [m]
        Discretized s-coordinates along drivetrain, measured from bedplate (direct) or tower center (geared)
    hss_diameter : numpy array[2], [m]
        Lss discretized diameter values at coordinates
    hss_wall_thickness : numpy array[2], [m]
        Lss discretized thickness values at coordinates
    M_hub : numpy array[3, n_dlcs], [N*m]
        Moment vector applied to the hub
    m_generator : float, [kg]
        Gearbox rotor mass
    cm_generator : float, [kg]
        Gearbox center of mass (measured from tower center)
    I_generator : numpy array[3], [kg*m**2]
        Gearbox moment of inertia (measured about its cm)
    hss_E : float, [Pa]
        modulus of elasticity
    hss_G : float, [Pa]
        shear modulus
    hss_rho : float, [kg/m**3]
        material density
    hss_Xy : float, [Pa]
        yield stress

    Returns
    -------
    hss_spring_constant : float, [N*m/rad]
        Equivalent spring constant for the high speed shaft froom T=(G*J/L)*theta
    hss_axial_stress : numpy array[5, n_dlcs], [Pa]
        Axial stress in Curved_beam structure
    hss_shear_stress : numpy array[5, n_dlcs], [Pa]
        Shear stress in Curved_beam structure
    hss_bending_stress : numpy array[5, n_dlcs], [Pa]
        Hoop stress in Curved_beam structure calculated with Roarks formulae
    constr_hss_vonmises : numpy array[5, n_dlcs]
        Sigma_y/Von_Mises
    F_generator : numpy array[3, n_dlcs], [N]
        Force vector applied to generator rotor (direct) or gearbox (geared) in hub c.s.
    M_generator : numpy array[3, n_dlcs], [N*m]
        Moment vector applied to generator rotor (direct) or gearbox (geared) in hub c.s.

    """

    def initialize(self):
        self.options.declare("n_dlcs")
        self.options.declare("modeling_options")

    def setup(self):
        n_dlcs = self.options["n_dlcs"]

        self.add_input("tilt", 0.0, units="deg")
        self.add_input("s_hss", val=np.zeros(3), units="m")
        self.add_input("hss_diameter", val=np.zeros(2), units="m")
        self.add_input("hss_wall_thickness", val=np.zeros(2), units="m")
        self.add_input("M_hub", val=np.zeros((3, n_dlcs)), units="N*m")
        self.add_input("gear_ratio", val=1.0)
        self.add_input("s_generator", val=0.0, units="m")
        self.add_input("generator_mass", val=0.0, units="kg")
        self.add_input("generator_I", val=np.zeros(3), units="kg*m**2")
        self.add_input("brake_mass", val=0.0, units="kg")
        self.add_input("brake_I", val=np.zeros(3), units="kg*m**2")
        self.add_input("hss_E", val=0.0, units="Pa")
        self.add_input("hss_G", val=0.0, units="Pa")
        self.add_input("hss_rho", val=0.0, units="kg/m**3")
        self.add_input("hss_Xy", val=0.0, units="Pa")

        self.add_output("hss_spring_constant", 0.0, units="N*m/rad")
        self.add_output("hss_axial_stress", np.zeros((2, n_dlcs)), units="Pa")
        self.add_output("hss_shear_stress", np.zeros((2, n_dlcs)), units="Pa")
        self.add_output("hss_bending_stress", np.zeros((2, n_dlcs)), units="Pa")
        self.add_output("constr_hss_vonmises", np.zeros((2, n_dlcs)))
        self.add_output("F_generator", val=np.zeros((3, n_dlcs)), units="N")
        self.add_output("M_generator", val=np.zeros((3, n_dlcs)), units="N*m")

    def compute(self, inputs, outputs):
        # Unpack inputs
        tilt = float(np.deg2rad(inputs["tilt"][0]))

        s_hss = inputs["s_hss"]
        D_hss = inputs["hss_diameter"]
        t_hss = inputs["hss_wall_thickness"]

        s_generator = float(inputs["s_generator"][0])
        m_generator = float(inputs["generator_mass"][0])
        I_generator = inputs["generator_I"]

        m_brake = float(inputs["brake_mass"][0])
        I_brake = inputs["brake_I"]

        rho = float(inputs["hss_rho"][0])
        E = float(inputs["hss_E"][0])
        G = float(inputs["hss_G"][0])
        sigma_y = float(inputs["hss_Xy"][0])
        gamma_f = float(self.options["modeling_options"]["gamma_f"])
        gamma_m = float(self.options["modeling_options"]["gamma_m"])
        gamma_n = float(self.options["modeling_options"]["gamma_n"])
        gamma = gamma_f * gamma_m * gamma_n

        M_hub = inputs["M_hub"]
        gear_ratio = float(inputs["gear_ratio"][0])

        # ------- node data ----------------
        n = len(s_hss)
        inode = np.arange(1, n + 1)
        ynode = znode = rnode = np.zeros(n)
        xnode = s_hss.copy()
        nodes = frame3dd.NodeData(inode, xnode, ynode, znode, rnode)
        # ------------------------------------

        # ------ reaction data ------------
        # Reaction at generator attachment
        rnode = [inode[0]]
        Rx = Ry = Rz = Rxx = Ryy = Rzz = np.array([RIGID])
        reactions = frame3dd.ReactionData(rnode, Rx, Ry, Rz, Rxx, Ryy, Rzz, rigid=RIGID)
        # -----------------------------------

        # ------ frame element data ------------
        hsscyl = tube_prop(s_hss, D_hss, t_hss)
        ielement = np.arange(1, n)
        N1 = np.arange(1, n)
        N2 = np.arange(2, n + 1)
        roll = np.zeros(n - 1)
        myones = np.ones(n - 1)
        Ax = hsscyl.Area
        As = hsscyl.Asx
        S = hsscyl.S
        C = hsscyl.C
        J0 = hsscyl.J0
        Jx = hsscyl.Ixx

        elements = frame3dd.ElementData(
            ielement, N1, N2, Ax, As, As, J0, Jx, Jx, E * myones, G * myones, roll, rho * myones
        )

        hss_stiff = G * J0 / np.diff(s_hss)
        outputs["hss_spring_constant"] = 1.0 / np.sum(1.0 / hss_stiff)
        # -----------------------------------

        # ------ options ------------
        shear = geom = True
        dx = -1
        options = frame3dd.Options(shear, geom, dx)
        # -----------------------------------

        # initialize frameDD3 object
        myframe = frame3dd.Frame(nodes, reactions, elements, options)

        # ------ add brake hub and generator rotor (direct) or generator (geared) extra mass ------------
        myframe.changeExtraNodeMass(
            np.r_[inode[1], inode[0]],
            [m_brake, m_generator],
            [I_brake[0], I_generator[0]],
            [I_brake[1], I_generator[1]],
            [I_brake[2], I_generator[2]],
            [0.0, 0.0],
            [0.0, 0.0],
            [0.0, 0.0],
            [0.0, s_generator - s_hss[-1]],
            [0.0, 0.0],
            [0.0, 0.0],
            True,
        )
        # ------------------------------------

        # ------ static load cases ------------
        n_dlcs = self.options["n_dlcs"]
        gy = 0.0
        gx = -gravity * np.sin(tilt)
        gz = -gravity * np.cos(tilt)
        for k in range(n_dlcs):
            # gravity in the X, Y, Z, directions (global)
            load = frame3dd.StaticLoadCase(gx, gy, gz)

            # point loads
            Fx = Fy = Fz = My = Mz = np.zeros(1)
            Mx = M_hub[0] / gear_ratio
            load.changePointLoads([inode[-1]], Fx, Fy, Fz, Mx, My, Mz)
            # -----------------------------------

            # Put all together and run
            myframe.addLoadCase(load)

        # myframe.write('myframe2.3dd') # Debugging
        displacements, forces, reactions, internalForces, mass3dd, modal = myframe.run()

        # Loop over DLCs and append to outputs
        outputs["F_generator"] = np.zeros((3, n_dlcs))
        outputs["M_generator"] = np.zeros((3, n_dlcs))
        outputs["hss_axial_stress"] = np.zeros((n - 1, n_dlcs))
        outputs["hss_shear_stress"] = np.zeros((n - 1, n_dlcs))
        outputs["hss_bending_stress"] = np.zeros((n - 1, n_dlcs))
        outputs["constr_hss_vonmises"] = np.zeros((n - 1, n_dlcs))
        for k in range(n_dlcs):
            # shear and bending, one per element (convert from local to global c.s.)
            Fx = forces.Nx[k, 1::2]
            Vy = forces.Vy[k, 1::2]
            Vz = -forces.Vz[k, 1::2]
            F = np.sqrt(Vz**2 + Vy**2)

            Mxx = forces.Txx[k, 1::2]
            Myy = forces.Myy[k, 1::2]
            Mzz = -forces.Mzz[k, 1::2]
            M = np.sqrt(Myy**2 + Mzz**2)

            # Record total forces and moments
            outputs["F_generator"][:, k] = -1.0 * np.array([reactions.Fx[k, 0], reactions.Fy[k, 0], reactions.Fz[k, 0]])
            outputs["M_generator"][:, k] = -1.0 * np.array(
                [reactions.Mxx[k, 0], reactions.Myy[k, 0], reactions.Mzz[k, 0]]
            )
            outputs["hss_axial_stress"][:, k] = np.abs(Fx) / Ax + M / S
            outputs["hss_shear_stress"][:, k] = 2.0 * F / As + np.abs(Mxx) / C
            hoop = np.zeros(F.shape)

            outputs["constr_hss_vonmises"][:, k] = TubevonMisesStressUtilization(
                outputs["hss_axial_stress"][:, k],
                hoop,
                outputs["hss_shear_stress"][:, k],
                gamma,
                sigma_y,
            )


class Nose_Stator_Bedplate_Frame(om.ExplicitComponent):
    """
    Run structural analysis of nose/turret with the generator stator and bedplate

    Parameters
    ----------
    upwind : boolean
        Flag whether the design is upwind or downwind
    tilt : float, [deg]
        Lss tilt
    s_nose : numpy array[5], [m]
        Discretized s-coordinates along drivetrain, measured from bedplate
    nose_diameter : numpy array[2], [m]
        Nose outer diameter from bearing 1 to bedplate
    nose_wall_thickness : numpy array[2], [m]
        Nose wall thickness
    x_bedplate : numpy array[12], [m]
        Bedplate centerline x-coordinates
    z_bedplate : numpy array[12], [m]
        Bedplate centerline z-coordinates
    x_bedplate_inner : numpy array[12], [m]
        Bedplate lower curve x-coordinates
    z_bedplate_inner : numpy array[12], [m]
        Bedplate lower curve z-coordinates
    x_bedplate_outer : numpy array[12], [m]
        Bedplate outer curve x-coordinates
    z_bedplate_outer : numpy array[12], [m]
        Bedplate outer curve z-coordinates
    D_bedplate : numpy array[12], [m]
        Bedplate diameters
    t_bedplate : numpy array[12], [m]
        Bedplate wall thickness (mirrors input)
    s_mb1 : float, [m]
        Bearing 1 s-coordinate along drivetrain, measured from bedplate
    s_mb2 : float, [m]
        Bearing 2 s-coordinate along drivetrain, measured from bedplate
    mb1_mass : float, [kg]
        component mass
    mb1_I : numpy array[3], [kg*m**2]
        component I
    mb1_max_defl_ang : float, [rad]
        Maximum allowable deflection angle
    mb2_mass : float, [kg]
        component mass
    mb2_I : numpy array[3], [kg*m**2]
        component I
    mb2_max_defl_ang : float, [rad]
        Maximum allowable deflection angle
    s_stator : float, [m]
        Generator stator attachment to lss s-coordinate measured from bedplate
    generator_stator_mass : float, [kg]
        Generator stator mass
    generator_stator_I : numpy array[3], [kg*m**2]
        Generator stator moment of inertia (measured about cm)
    F_mb1 : numpy array[3, n_dlcs], [N]
        Force vector applied to bearing 1 in hub c.s.
    F_mb2 : numpy array[3, n_dlcs], [N]
        Force vector applied to bearing 2 in hub c.s.
    M_mb1 : numpy array[3, n_dlcs], [N*m]
        Moment vector applied to bearing 1 in hub c.s.
    M_mb2 : numpy array[3, n_dlcs], [N*m]
        Moment vector applied to bearing 2 in hub c.s.
    other_mass : float, [kg]
        Mass of other nacelle components that rest on mainplate
    bedplate_E : float, [Pa]
        modulus of elasticity
    bedplate_G : float, [Pa]
        shear modulus
    bedplate_rho : float, [kg/m**3]
        material density
    bedplate_Xy : float, [Pa]
        yield stress

    Returns
    -------
    mb1_deflection : numpy array[n_dlcs], [m]
        Total deflection distance of bearing 1
    mb2_deflection : numpy array[n_dlcs], [m]
        Total deflection distance of bearing 2
    stator_deflection : float, [m]
        Maximum deflection distance at stator attachment
    mb1_angle : numpy array[n_dlcs], [rad]
        Total rotation angle of bearing 1
    mb2_angle : numpy array[n_dlcs], [rad]
        Total rotation angle of bearing 2
    stator_angle : float, [rad]
        Maximum rotation angle at stator attachment
    base_F : numpy array[3, n_dlcs], [N]
        Total reaction force at bedplate base in tower top coordinate system
    base_M : numpy array[3, n_dlcs], [N*m]
        Total reaction moment at bedplate base in tower top coordinate system
    bedplate_nose_axial_stress : numpy array[12+3, n_dlcs], [Pa]
        Axial stress in Curved_beam structure
    bedplate_nose_shear_stress : numpy array[12+3, n_dlcs], [Pa]
        Shear stress in Curved_beam structure
    bedplate_nose_bending_stress : numpy array[12+3, n_dlcs], [Pa]
        Hoop stress in Curved_beam structure calculated with Roarks formulae
    constr_bedplate_vonmises : numpy array[12+3, n_dlcs]
        Sigma_y/Von_Mises
    constr_mb1_defl : numpy array[n_dlcs]
        Angular deflection relative to limit of bearing 1 (should be <1)
    constr_mb2_defl : numpy array[n_dlcs]
        Angular deflection relative to limit of bearing 2 (should be <1)

    """

    def initialize(self):
        self.options.declare("n_dlcs")
        self.options.declare("modeling_options")

    def setup(self):
        n_dlcs = self.options["n_dlcs"]

        self.add_discrete_input("upwind", True)
        self.add_input("tilt", 0.0, units="deg")
        self.add_input("s_nose", val=np.zeros(5), units="m")
        self.add_input("nose_diameter", np.zeros(2), units="m")
        self.add_input("nose_wall_thickness", np.zeros(2), units="m")
        self.add_input("x_bedplate", val=np.zeros(12), units="m")
        self.add_input("z_bedplate", val=np.zeros(12), units="m")
        self.add_input("x_bedplate_inner", val=np.zeros(12), units="m")
        self.add_input("z_bedplate_inner", val=np.zeros(12), units="m")
        self.add_input("x_bedplate_outer", val=np.zeros(12), units="m")
        self.add_input("z_bedplate_outer", val=np.zeros(12), units="m")
        self.add_input("D_bedplate", val=np.zeros(12), units="m")
        self.add_input("t_bedplate", val=np.zeros(12), units="m")
        self.add_input("s_mb1", val=0.0, units="m")
        self.add_input("s_mb2", val=0.0, units="m")
        self.add_input("mb1_mass", 0.0, units="kg")
        self.add_input("mb1_I", np.zeros(3), units="kg*m**2")
        self.add_input("mb1_max_defl_ang", 0.0, units="rad")
        self.add_input("mb2_mass", 0.0, units="kg")
        self.add_input("mb2_I", np.zeros(3), units="kg*m**2")
        self.add_input("mb2_max_defl_ang", 0.0, units="rad")
        self.add_input("s_stator", val=0.0, units="m")
        self.add_input("generator_stator_mass", val=0.0, units="kg")
        self.add_input("generator_stator_I", val=np.zeros(3), units="kg*m**2")
        self.add_input("F_mb1", val=np.zeros((3, n_dlcs)), units="N")
        self.add_input("F_mb2", val=np.zeros((3, n_dlcs)), units="N")
        self.add_input("M_mb1", val=np.zeros((3, n_dlcs)), units="N*m")
        self.add_input("M_mb2", val=np.zeros((3, n_dlcs)), units="N*m")
        self.add_input("other_mass", val=0.0, units="kg")
        self.add_input("bedplate_E", val=0.0, units="Pa")
        self.add_input("bedplate_G", val=0.0, units="Pa")
        self.add_input("bedplate_rho", val=0.0, units="kg/m**3")
        self.add_input("bedplate_Xy", val=0.0, units="Pa")

        self.add_input("stator_deflection_allowable", val=1.0, units="m")
        self.add_input("stator_angle_allowable", val=1.0, units="rad")

        self.add_output("mb1_deflection", val=np.zeros(n_dlcs), units="m")
        self.add_output("mb2_deflection", val=np.zeros(n_dlcs), units="m")
        self.add_output("stator_deflection", val=0.0, units="m")
        self.add_output("mb1_angle", val=np.zeros(n_dlcs), units="rad")
        self.add_output("mb2_angle", val=np.zeros(n_dlcs), units="rad")
        self.add_output("stator_angle", val=0.0, units="rad")
        self.add_output("base_F", val=np.zeros((3, n_dlcs)), units="N")
        self.add_output("base_M", val=np.zeros((3, n_dlcs)), units="N*m")
        self.add_output("bedplate_nose_axial_stress", np.zeros((12 + 3, n_dlcs)), units="Pa")
        self.add_output("bedplate_nose_shear_stress", np.zeros((12 + 3, n_dlcs)), units="Pa")
        self.add_output("bedplate_nose_bending_stress", np.zeros((12 + 3, n_dlcs)), units="Pa")
        self.add_output("constr_bedplate_vonmises", np.zeros((12 + 3, n_dlcs)))
        self.add_output("constr_mb1_defl", val=np.zeros(n_dlcs))
        self.add_output("constr_mb2_defl", val=np.zeros(n_dlcs))
        self.add_output("constr_stator_deflection", 0.0)
        self.add_output("constr_stator_angle", 0.0)

    def compute(self, inputs, outputs, discrete_inputs, discrete_outputs):
        # Unpack inputs
        upwind = discrete_inputs["upwind"]
        Cup = -1.0 if upwind else 1.0
        tiltD = float(inputs["tilt"][0])
        tiltR = np.deg2rad(tiltD)

        x_c = inputs["x_bedplate"]
        z_c = inputs["z_bedplate"]
        x_inner = inputs["x_bedplate_inner"]
        z_inner = inputs["z_bedplate_inner"]
        x_outer = inputs["x_bedplate_outer"]
        z_outer = inputs["z_bedplate_outer"]
        D_bed = inputs["D_bedplate"]
        t_bed = inputs["t_bedplate"]

        s_nose = inputs["s_nose"][1:]  # First point duplicated with bedplate
        D_nose = inputs["nose_diameter"]
        t_nose = inputs["nose_wall_thickness"]
        x_nose = s_nose.copy()
        x_nose *= Cup
        x_nose += x_c[-1]

        s_mb1 = float(inputs["s_mb1"][0])
        s_mb2 = float(inputs["s_mb2"][0])
        m_mb1 = float(inputs["mb1_mass"][0])
        m_mb2 = float(inputs["mb2_mass"][0])
        I_mb1 = inputs["mb1_I"]
        I_mb2 = inputs["mb2_I"]

        s_stator = float(inputs["s_stator"][0])
        m_stator = float(inputs["generator_stator_mass"][0])
        I_stator = inputs["generator_stator_I"]

        rho = float(inputs["bedplate_rho"][0])
        E = float(inputs["bedplate_E"][0])
        G = float(inputs["bedplate_G"][0])
        sigma_y = float(inputs["bedplate_Xy"][0])
        gamma_f = float(self.options["modeling_options"]["gamma_f"])
        gamma_m = float(self.options["modeling_options"]["gamma_m"])
        gamma_n = float(self.options["modeling_options"]["gamma_n"])
        gamma = gamma_f * gamma_m * gamma_n

        F_mb1 = inputs["F_mb1"]
        F_mb2 = inputs["F_mb2"]
        M_mb1 = inputs["M_mb1"]
        M_mb2 = inputs["M_mb2"]

        m_other = float(inputs["other_mass"][0])
        stator_defl_allow = float(inputs["stator_deflection_allowable"][0])
        stator_angle_allow = float(inputs["stator_angle_allowable"][0])

        # ------- node data ----------------
        n = len(x_c) + len(x_nose)
        inode = np.arange(1, n + 1)
        ynode = rnode = np.zeros(n)
        xnode = np.r_[x_c, x_nose]
        znode = np.r_[z_c, z_c[-1] * np.ones(x_nose.shape)]
        nodes = frame3dd.NodeData(inode, xnode, ynode, znode, rnode)
        # Grab indices for later
        inose = len(x_c)
        istator = inode[find_nearest(xnode, Cup * s_stator + x_c[-1])]
        i1 = inode[find_nearest(xnode, Cup * s_mb1 + x_c[-1])]
        i2 = inode[find_nearest(xnode, Cup * s_mb2 + x_c[-1])]
        # ------------------------------------

        # ------ reaction data ------------
        # Rigid base
        rnode = [int(inode[0])]
        rk = np.array([RIGID])
        reactions = frame3dd.ReactionData(rnode, rk, rk, rk, rk, rk, rk, rigid=RIGID)
        # -----------------------------------

        # ------ frame element data ------------
        bedcyl = tube_prop(x_c, D_bed, t_bed)
        nosecyl = tube_prop(inputs["s_nose"], D_nose, t_nose)
        ielement = np.arange(1, n)
        N1 = np.arange(1, n)
        N2 = np.arange(2, n + 1)
        roll = np.zeros(n - 1)
        myones = np.ones(n - 1)
        Ax = np.r_[bedcyl.Area, nosecyl.Area]
        As = np.r_[bedcyl.Asx, nosecyl.Asx]
        S = np.r_[bedcyl.S, nosecyl.S]
        C = np.r_[bedcyl.C, nosecyl.C]
        J0 = np.r_[bedcyl.J0, nosecyl.J0]
        Ixx = np.r_[bedcyl.Ixx, nosecyl.Ixx]

        elements = frame3dd.ElementData(
            ielement, N1, N2, Ax, As, As, J0, Ixx, Ixx, E * myones, G * myones, roll, rho * myones
        )
        # -----------------------------------

        # ------ options ------------
        shear = geom = True
        dx = -1
        options = frame3dd.Options(shear, geom, dx)
        # -----------------------------------

        # initialize frameDD3 object
        myframe = frame3dd.Frame(nodes, reactions, elements, options)

        # ------ add misc nacelle components at base and stator extra mass ------------
        myframe.changeExtraNodeMass(
            np.r_[inode[0], istator, i1, i2],
            [m_other, m_stator, m_mb1, m_mb2],
            [0.0, I_stator[0], I_mb1[0], I_mb2[0]],
            [0.0, I_stator[1], I_mb1[1], I_mb2[1]],
            [0.0, I_stator[2], I_mb1[2], I_mb2[2]],
            np.zeros(4),
            np.zeros(4),
            np.zeros(4),
            [0.0, 0.0, 0.0, 0.0],
            np.zeros(4),
            np.zeros(4),
            True,
        )
        # ------------------------------------

        # ------- NO dynamic analysis ----------
        # myframe.enableDynamics(NFREQ, discrete_inputs['Mmethod'], discrete_inputs['lump'], float(inputs['tol']), float(inputs['shift']))
        # ----------------------------

        # ------ static load cases ------------
        n_dlcs = self.options["n_dlcs"]
        gy = 0.0
        gx = -gravity * np.sin(tiltR)
        gz = -gravity * np.cos(tiltR)
        for k in range(n_dlcs):
            # gravity in the X, Y, Z, directions (global)
            load = frame3dd.StaticLoadCase(gx, gy, gz)

            # point loads
            F_12 = np.c_[F_mb2[:, k], F_mb1[:, k]]
            M_12 = np.c_[M_mb2[:, k], M_mb1[:, k]]
            load.changePointLoads(np.r_[i2, i1], F_12[0, :], F_12[1, :], F_12[2, :], M_12[0, :], M_12[1, :], M_12[2, :])
            # -----------------------------------

            # Put all together and run
            myframe.addLoadCase(load)

        # myframe.write('myframe3.3dd') # Debugging
        displacements, forces, reactions, internalForces, mass3dd, modal = myframe.run()

        # ------------ Bedplate "curved beam" geometry for post-processing -------------
        # Need to compute neutral axis, so shift points such that bedplate top is at x=0
        R_c = np.sqrt((x_c - x_c[-1]) ** 2 + z_c**2)
        Ro = np.sqrt((x_outer - x_c[-1]) ** 2 + z_outer**2)
        Ri = np.sqrt((x_inner - x_c[-1]) ** 2 + z_inner**2)
        r_bed_o = 0.5 * D_bed
        r_bed_i = r_bed_o - t_bed
        A_bed = np.pi * (r_bed_o**2 - r_bed_i**2)

        # Radius of the neutral axis
        # http://faculty.fairfield.edu/wdornfeld/ME311/BasicStressEqns-DBWallace.pdf
        R_n = A_bed / (2 * np.pi * (np.sqrt(R_c**2 - r_bed_i**2) - np.sqrt(R_c**2 - r_bed_o**2)))
        e_cn = R_c - R_n
        # ------------------------------------

        # Loop over DLCs and append to outputs
        outputs["mb1_deflection"] = np.zeros(n_dlcs)
        outputs["mb2_deflection"] = np.zeros(n_dlcs)
        stator_deflection = np.zeros(n_dlcs)
        outputs["mb1_angle"] = np.zeros(n_dlcs)
        outputs["mb2_angle"] = np.zeros(n_dlcs)
        stator_angle = np.zeros(n_dlcs)
        outputs["base_F"] = np.zeros((3, n_dlcs))
        outputs["base_M"] = np.zeros((3, n_dlcs))
        outputs["bedplate_nose_axial_stress"] = np.zeros((n - 1, n_dlcs))
        outputs["bedplate_nose_shear_stress"] = np.zeros((n - 1, n_dlcs))
        outputs["bedplate_nose_bending_stress"] = np.zeros((n - 1, n_dlcs))
        outputs["constr_bedplate_vonmises"] = np.zeros((n - 1, n_dlcs))
        for k in range(n_dlcs):
            # Deflections and rotations at bearings- how to sum up rotation angles?
            outputs["mb1_deflection"][k] = np.sqrt(
                displacements.dx[k, i1 - 1] ** 2 + displacements.dy[k, i1 - 1] ** 2 + displacements.dz[k, i1 - 1] ** 2
            )
            outputs["mb2_deflection"][k] = np.sqrt(
                displacements.dx[k, i2 - 1] ** 2 + displacements.dy[k, i2 - 1] ** 2 + displacements.dz[k, i2 - 1] ** 2
            )
            stator_deflection[k] = np.sqrt(
                displacements.dx[k, istator - 1] ** 2
                + displacements.dy[k, istator - 1] ** 2
                + displacements.dz[k, istator - 1] ** 2
            )
            outputs["mb1_angle"][k] = (
                displacements.dxrot[k, i1 - 1] + displacements.dyrot[k, i1 - 1] + displacements.dzrot[k, i1 - 1]
            )
            outputs["mb2_angle"][k] = (
                displacements.dxrot[k, i2 - 1] + displacements.dyrot[k, i2 - 1] + displacements.dzrot[k, i2 - 1]
            )
            stator_angle[k] = (
                displacements.dxrot[k, istator - 1]
                + displacements.dyrot[k, istator - 1]
                + displacements.dzrot[k, istator - 1]
            )

            # shear and bending, one per element (convert from local to global c.s.)
            Fx = forces.Nx[k, 1::2]
            Vy = forces.Vy[k, 1::2]
            Vz = -forces.Vz[k, 1::2]
            F = np.sqrt(Vz**2 + Vy**2)

            Mxx = forces.Txx[k, 1::2]
            Myy = forces.Myy[k, 1::2]
            Mzz = -forces.Mzz[k, 1::2]
            M = np.sqrt(Myy**2 + Mzz**2)

            # Record total forces and moments
            F_base_k = DirectionVector(-reactions.Fx[k, :].sum(), -reactions.Fy[k, :].sum(), -reactions.Fz[k, :].sum())
            M_base_k = DirectionVector(
                -reactions.Mxx[k, :].sum(), -reactions.Myy[k, :].sum(), -reactions.Mzz[k, :].sum()
            )

            # Rotate vector from tilt axes to yaw/tower axes
            outputs["base_F"][:, k] = F_base_k.hubToYaw(-tiltD).toArray()
            outputs["base_M"][:, k] = M_base_k.hubToYaw(-tiltD).toArray()

            outputs["bedplate_nose_axial_stress"][:, k] = np.abs(Fx) / Ax + M / S
            outputs["bedplate_nose_shear_stress"][:, k] = 2.0 * F / As + np.abs(Mxx) / C

            Bending_stress_outer = M[: (inose - 1)] * nodal2sectional((Ro - R_n) / (A_bed * e_cn * Ro))[0]
            # Bending_stress_inner = M[:(inose-1)] * nodal2sectional( (R_n-Ri) / (A_bed*e_cn*Ri) )[0]
            outputs["bedplate_nose_bending_stress"][: (inose - 1), k] = Bending_stress_outer

            outputs["constr_bedplate_vonmises"][:, k] = TubevonMisesStressUtilization(
                outputs["bedplate_nose_axial_stress"][:, k],
                outputs["bedplate_nose_bending_stress"][:, k],
                outputs["bedplate_nose_shear_stress"][:, k],
                gamma,
                sigma_y,
            )

        # Evaluate bearing limits
        outputs["constr_mb1_defl"] = np.abs(outputs["mb1_angle"] / inputs["mb1_max_defl_ang"])
        outputs["constr_mb2_defl"] = np.abs(outputs["mb2_angle"] / inputs["mb2_max_defl_ang"])
        outputs["stator_deflection"] = stator_deflection.max()
        outputs["stator_angle"] = np.abs(stator_angle).max()
        outputs["constr_stator_deflection"] = gamma * outputs["stator_deflection"] / stator_defl_allow
        outputs["constr_stator_angle"] = gamma * outputs["stator_angle"] / stator_angle_allow


class Bedplate_IBeam_Frame(om.ExplicitComponent):
    """
    Run structural analysis of bedplate in geared configuration with 2 X 2 parallel I beams providing support

    Parameters
    ----------
    upwind : boolean
        Flag whether the design is upwind or downwind
    tilt : float, [deg]
        Lss tilt
    bedplate_flange_width : float, [m]
        Bedplate is two parallel I beams, this is the flange width
    bedplate_flange_thickness : float, [m]
        Bedplate is two parallel I beams, this is the flange thickness
    bedplate_web_thickness : float, [m]
        Bedplate is two parallel I beams, this is the web thickness
    bedplate_web_height : float, [m]
        Bedplate is two parallel I beams, this is the web height
    s_mb1 : float, [m]
        Bearing 1 s-coordinate along drivetrain, measured from bedplate
    s_mb2 : float, [m]
        Bearing 2 s-coordinate along drivetrain, measured from bedplate
    mb1_mass : float, [kg]
        component mass
    mb1_I : numpy array[3], [kg*m**2]
        component I
    mb1_max_defl_ang : float, [rad]
        Maximum allowable deflection angle
    mb2_mass : float, [kg]
        component mass
    mb2_I : numpy array[3], [kg*m**2]
        component I
    mb2_max_defl_ang : float, [rad]
        Maximum allowable deflection angle
    F_mb1 : numpy array[3, n_dlcs], [N]
        Force vector applied to bearing 1 in hub c.s.
    F_mb2 : numpy array[3, n_dlcs], [N]
        Force vector applied to bearing 2 in hub c.s.
    M_mb1 : numpy array[3, n_dlcs], [N*m]
        Moment vector applied to bearing 1 in hub c.s.
    M_mb2 : numpy array[3, n_dlcs], [N*m]
        Moment vector applied to bearing 2 in hub c.s.
    other_mass : float, [kg]
        Mass of other nacelle components that rest on mainplate
    E : float, [Pa]
        modulus of elasticity
    G : float, [Pa]
        shear modulus
    rho : float, [kg/m**3]
        material density
    sigma_y : float, [Pa]
        yield stress

    Returns
    -------
    mb1_deflection : numpy array[n_dlcs], [m]
        Total deflection distance of bearing 1
    mb2_deflection : numpy array[n_dlcs], [m]
        Total deflection distance of bearing 2
    bedplate_deflection : float, [m]
        Maximum deflection distance at bedplate I-beam tip
    mb1_angle : numpy array[n_dlcs], [rad]
        Total rotation angle of bearing 1
    mb2_angle : numpy array[n_dlcs], [rad]
        Total rotation angle of bearing 2
    bedplate_angle : float, [rad]
        Maximum rotation angle at bedplate I-beam tip
    base_F : numpy array[3, n_dlcs], [N]
        Total reaction force at bedplate base in tower top coordinate system
    base_M : numpy array[3, n_dlcs], [N*m]
        Total reaction moment at bedplate base in tower top coordinate system
    bedplate_axial_stress : numpy array[22, n_dlcs], [Pa]
        Axial stress in Curved_beam structure
    bedplate_shear_stress : numpy array[22, n_dlcs], [Pa]
        Shear stress in Curved_beam structure
    bedplate_bending_stress : numpy array[22, n_dlcs], [Pa]
        Hoop stress in Curved_beam structure calculated with Roarks formulae
    constr_bedplate_vonmises : numpy array[22, n_dlcs]
        Sigma_y/Von_Mises
    constr_mb1_defl : numpy array[n_dlcs]
        Angular deflection relative to limit of bearing 1 (should be <1)
    constr_mb2_defl : numpy array[n_dlcs]
        Angular deflection relative to limit of bearing 2 (should be <1)

    """

    def initialize(self):
        self.options.declare("n_dlcs")
        self.options.declare("modeling_options")

    def setup(self):
        n_dlcs = self.options["n_dlcs"]

        self.add_discrete_input("upwind", True)
        self.add_input("tilt", 0.0, units="deg")
        self.add_input("D_top", 0.0, units="m")
        self.add_input("s_drive", val=np.zeros(12), units="m")
        self.add_input("bedplate_flange_width", val=0.0, units="m")
        self.add_input("bedplate_flange_thickness", val=0.0, units="m")
        self.add_input("bedplate_web_height", val=0.0, units="m")
        self.add_input("bedplate_web_thickness", val=0.0, units="m")
        self.add_input("s_mb1", val=0.0, units="m")
        self.add_input("s_mb2", val=0.0, units="m")
        self.add_input("mb1_mass", 0.0, units="kg")
        self.add_input("mb1_I", np.zeros(3), units="kg*m**2")
        self.add_input("mb1_max_defl_ang", 0.0, units="rad")
        self.add_input("mb2_mass", 0.0, units="kg")
        self.add_input("mb2_I", np.zeros(3), units="kg*m**2")
        self.add_input("mb2_max_defl_ang", 0.0, units="rad")
        self.add_input("s_gearbox", val=0.0, units="m")
        self.add_input("s_generator", val=0.0, units="m")
        self.add_input("F_mb1", val=np.zeros((3, n_dlcs)), units="N")
        self.add_input("F_mb2", val=np.zeros((3, n_dlcs)), units="N")
        self.add_input("F_torq", val=np.zeros((3, n_dlcs)), units="N")
        self.add_input("F_generator", val=np.zeros((3, n_dlcs)), units="N")
        self.add_input("M_mb1", val=np.zeros((3, n_dlcs)), units="N*m")
        self.add_input("M_mb2", val=np.zeros((3, n_dlcs)), units="N*m")
        self.add_input("M_torq", val=np.zeros((3, n_dlcs)), units="N*m")
        self.add_input("M_generator", val=np.zeros((3, n_dlcs)), units="N*m")
        self.add_input("other_mass", val=0.0, units="kg")
        self.add_input("bedplate_E", val=0.0, units="Pa")
        self.add_input("bedplate_G", val=0.0, units="Pa")
        self.add_input("bedplate_rho", val=0.0, units="kg/m**3")
        self.add_input("bedplate_Xy", val=0.0, units="Pa")

        self.add_input("stator_deflection_allowable", val=1.0, units="m")
        self.add_input("stator_angle_allowable", val=1.0, units="rad")

        self.add_output("mb1_deflection", val=np.zeros(n_dlcs), units="m")
        self.add_output("mb2_deflection", val=np.zeros(n_dlcs), units="m")
        self.add_output("stator_deflection", val=0.0, units="m")
        self.add_output("mb1_angle", val=np.zeros(n_dlcs), units="rad")
        self.add_output("mb2_angle", val=np.zeros(n_dlcs), units="rad")
        self.add_output("stator_angle", val=0.0, units="rad")
        self.add_output("base_F", val=np.zeros((3, n_dlcs)), units="N")
        self.add_output("base_M", val=np.zeros((3, n_dlcs)), units="N*m")
        self.add_output("bedplate_axial_stress", np.zeros((22, n_dlcs)), units="Pa")
        self.add_output("bedplate_shear_stress", np.zeros((22, n_dlcs)), units="Pa")
        self.add_output("bedplate_bending_stress", np.zeros((22, n_dlcs)), units="Pa")
        self.add_output("constr_bedplate_vonmises", np.zeros((22, n_dlcs)))
        self.add_output("constr_mb1_defl", val=np.zeros(n_dlcs))
        self.add_output("constr_mb2_defl", val=np.zeros(n_dlcs))
        self.add_output("constr_stator_deflection", 0.0)
        self.add_output("constr_stator_angle", 0.0)

    def compute(self, inputs, outputs, discrete_inputs, discrete_outputs):
        # Unpack inputs
        upwind = discrete_inputs["upwind"]
        Cup = -1.0 if upwind else 1.0
        D_top = float(inputs["D_top"][0])
        tiltD = float(inputs["tilt"][0])
        tiltR = np.deg2rad(tiltD)
        s_drive = inputs["s_drive"]

        s_gear = float(inputs["s_gearbox"][0])
        s_gen = float(inputs["s_generator"][0])

        s_mb1 = float(inputs["s_mb1"][0])
        s_mb2 = float(inputs["s_mb2"][0])
        m_mb1 = float(inputs["mb1_mass"][0])
        m_mb2 = float(inputs["mb2_mass"][0])
        I_mb1 = inputs["mb1_I"]
        I_mb2 = inputs["mb2_I"]

        bed_w_flange = float(inputs["bedplate_flange_width"][0])
        bed_t_flange = float(inputs["bedplate_flange_thickness"][0])
        bed_h_web = float(inputs["bedplate_web_height"][0])
        bed_t_web = float(inputs["bedplate_web_thickness"][0])

        rho = float(inputs["bedplate_rho"][0])
        E = float(inputs["bedplate_E"][0])
        G = float(inputs["bedplate_G"][0])
        sigma_y = float(inputs["bedplate_Xy"][0])
        gamma_f = float(self.options["modeling_options"]["gamma_f"])
        gamma_m = float(self.options["modeling_options"]["gamma_m"])
        gamma_n = float(self.options["modeling_options"]["gamma_n"])
        gamma = gamma_f * gamma_m * gamma_n

        F_mb1 = inputs["F_mb1"]
        F_mb2 = inputs["F_mb2"]
        F_gear = inputs["F_torq"]
        F_gen = inputs["F_generator"]
        M_mb1 = inputs["M_mb1"]
        M_mb2 = inputs["M_mb2"]
        M_gear = inputs["M_torq"]
        M_gen = inputs["M_generator"]

        m_other = float(inputs["other_mass"][0])
        stator_defl_allow = float(inputs["stator_deflection_allowable"][0])
        stator_angle_allow = float(inputs["stator_angle_allowable"][0])

        # ------- node data ----------------
        n = len(s_drive)
        inode = np.arange(1, 3 * n + 1)
        ynode = 0.25 * D_top * np.r_[np.zeros(n), np.ones(n), -np.ones(n)]
        xnode = s_drive * np.cos(tiltR)
        xnode = Cup * np.r_[xnode, xnode, xnode]
        znode = rnode = np.zeros(3 * n)
        nodes = frame3dd.NodeData(inode, xnode, ynode, znode, rnode)
        # Grab indices for later
        igenerator = inode[find_nearest(s_drive, s_gen)]
        igearbox = inode[find_nearest(s_drive, s_gear)]
        itower = inode[find_nearest(s_drive, 0.0)]
        i1 = inode[find_nearest(s_drive, s_mb1)]
        i2 = inode[find_nearest(s_drive, s_mb2)]
        # ------------------------------------

        # ------ reaction data ------------
        # Rigid base
        rnode = np.int_(np.r_[itower + n, itower + 2 * n])
        rk = np.array([RIGID, RIGID])
        reactions = frame3dd.ReactionData(rnode, rk, rk, rk, rk, rk, rk, rigid=RIGID)
        # -----------------------------------

        # ------ frame element data ------------
        # 2 parallel I-beams, with inner line of points to receive loads & distribute
        # No connections on ghost nodes, just connections to I-beams at same xval
        # Elements connect the lines, then connect the I-beams to the center nodes
        myI = IBeam(bed_w_flange, bed_t_flange, bed_h_web, bed_t_web)

        N1i = np.arange(1, n)
        N2i = np.arange(2, n + 1)
        Ni = np.arange(1, n + 1)
        N1 = np.r_[N1i + n, N1i + 2 * n, Ni + n, Ni + 2 * n]
        N2 = np.r_[N2i + n, N2i + 2 * n, Ni, Ni]
        ielement = np.arange(1, N1.size + 1)
        roll = np.zeros(N1.size)
        myones = np.ones(N1i.size)
        plate1s = np.ones(Ni.size)

        A_plate = As_plate = S_plate = C_plate = I_plate = 1e3 * plate1s
        rho_plate = 1e-6 * plate1s
        E_plate = G_plate = 1e16 * plate1s

        Ax = np.r_[myI.Area * myones, myI.Area * myones, A_plate, A_plate]
        Asy = np.r_[myI.Asx * myones, myI.Asx * myones, As_plate, As_plate]
        Asz = np.r_[myI.Asy * myones, myI.Asy * myones, As_plate, As_plate]
        Sy = np.r_[myI.Sx * myones, myI.Sx * myones, S_plate, S_plate]
        Sz = np.r_[myI.Sy * myones, myI.Sy * myones, S_plate, S_plate]
        C = np.r_[myI.C * myones, myI.C * myones, C_plate, C_plate]
        J0 = np.r_[myI.J0 * myones, myI.J0 * myones, I_plate, I_plate]
        Jy = np.r_[myI.Ixx * myones, myI.Ixx * myones, I_plate, I_plate]
        Jz = np.r_[myI.Iyy * myones, myI.Iyy * myones, I_plate, I_plate]
        myE = np.r_[E * myones, E * myones, E_plate, E_plate]
        myG = np.r_[G * myones, G * myones, G_plate, G_plate]
        myrho = np.r_[rho * myones, rho * myones, rho_plate, rho_plate]

        elements = frame3dd.ElementData(ielement, N1, N2, Ax, Asy, Asz, J0, Jy, Jz, myE, myG, roll, myrho)
        # -----------------------------------

        # ------ options ------------
        shear = geom = True
        dx = -1
        options = frame3dd.Options(shear, geom, dx)
        # -----------------------------------

        # initialize frameDD3 object
        myframe = frame3dd.Frame(nodes, reactions, elements, options)

        # ------ add misc nacelle components at base and stator extra mass ------------
        myframe.changeExtraNodeMass(
            np.r_[itower, i1, i2],
            [m_other, m_mb1, m_mb2],
            [0.0, I_mb1[0], I_mb2[0]],
            [0.0, I_mb1[1], I_mb2[1]],
            [0.0, I_mb1[2], I_mb2[2]],
            np.zeros(3),
            np.zeros(3),
            np.zeros(3),
            np.zeros(3),
            np.zeros(3),
            np.zeros(3),
            True,
        )
        # ------------------------------------

        # ------- NO dynamic analysis ----------
        # myframe.enableDynamics(NFREQ, discrete_inputs['Mmethod'], discrete_inputs['lump'], float(inputs['tol']), float(inputs['shift']))
        # ----------------------------

        # ------ static load cases ------------
        n_dlcs = self.options["n_dlcs"]
        gx = gy = 0.0
        gz = -gravity
        for k in range(n_dlcs):
            # gravity in the X, Y, Z, directions (global)
            load = frame3dd.StaticLoadCase(gx, gy, gz)

            # point loads
            F_ext = np.c_[F_mb2[:, k], F_mb1[:, k], F_gear[:, k], F_gen[:, k]]
            M_ext = np.c_[M_mb2[:, k], M_mb1[:, k], M_gear[:, k], M_gen[:, k]]
            F_rot = DirectionVector(F_ext[0, :], F_ext[1, :], F_ext[2, :]).hubToYaw(-tiltD).toArray()
            M_rot = DirectionVector(M_ext[0, :], M_ext[1, :], M_ext[2, :]).hubToYaw(-tiltD).toArray()
            load.changePointLoads(
                np.r_[i2, i1, igearbox, igenerator],
                F_rot[:, 0],
                F_rot[:, 1],
                F_rot[:, 2],
                M_rot[:, 0],
                M_rot[:, 1],
                M_rot[:, 2],
            )
            # -----------------------------------

            # Put all together and run
            myframe.addLoadCase(load)

        # myframe.write('myframe4.3dd') # Debugging
        displacements, forces, reactions, internalForces, mass3dd, modal = myframe.run()

        # Loop over DLCs and append to outputs
        outputs["mb1_deflection"] = np.zeros(n_dlcs)
        outputs["mb2_deflection"] = np.zeros(n_dlcs)
        outputs["mb1_angle"] = np.zeros(n_dlcs)
        outputs["mb2_angle"] = np.zeros(n_dlcs)
        outputs["base_F"] = np.zeros((3, n_dlcs))
        outputs["base_M"] = np.zeros((3, n_dlcs))
        outputs["bedplate_axial_stress"] = np.zeros((2 * n - 2, n_dlcs))
        outputs["bedplate_shear_stress"] = np.zeros((2 * n - 2, n_dlcs))
        outputs["bedplate_bending_stress"] = np.zeros((2 * n - 2, n_dlcs))
        outputs["constr_bedplate_vonmises"] = np.zeros((2 * n - 2, n_dlcs))
        for k in range(n_dlcs):
            # Deflections and rotations at bearings- how to sum up rotation angles?
            outputs["mb1_deflection"][k] = np.sqrt(
                displacements.dx[k, i1 - 1] ** 2 + displacements.dy[k, i1 - 1] ** 2 + displacements.dz[k, i1 - 1] ** 2
            )
            outputs["mb2_deflection"][k] = np.sqrt(
                displacements.dx[k, i2 - 1] ** 2 + displacements.dy[k, i2 - 1] ** 2 + displacements.dz[k, i2 - 1] ** 2
            )
            bedplate_deflection = np.maximum(
                np.sqrt(displacements.dx[k, n] ** 2 + displacements.dy[k, n] ** 2 + displacements.dz[k, n] ** 2),
                np.sqrt(displacements.dx[k, -1] ** 2 + displacements.dy[k, -1] ** 2 + displacements.dz[k, -1] ** 2),
            )
            outputs["mb1_angle"][k] = (
                displacements.dxrot[k, i1 - 1] + displacements.dyrot[k, i1 - 1] + displacements.dzrot[k, i1 - 1]
            )
            outputs["mb2_angle"][k] = (
                displacements.dxrot[k, i2 - 1] + displacements.dyrot[k, i2 - 1] + displacements.dzrot[k, i2 - 1]
            )
            bedplate_angle = np.maximum(
                displacements.dxrot[k, n] + displacements.dyrot[k, n] + displacements.dzrot[k, n],
                displacements.dxrot[k, -1] + displacements.dyrot[k, -1] + displacements.dzrot[k, -1],
            )

            # shear and bending, one per element (convert from local to global c.s.)
            Fx = forces.Nx[k, 1::2]
            Vy = forces.Vy[k, 1::2]
            Vz = -forces.Vz[k, 1::2]
            # F  =  np.sqrt(Vz**2 + Vy**2)

            Mxx = forces.Txx[k, 1::2]
            Myy = forces.Myy[k, 1::2]
            Mzz = -forces.Mzz[k, 1::2]
            # M   =  np.sqrt(Myy**2 + Mzz**2)

            # Record total forces and moments at base
            outputs["base_F"][:, k] = np.r_[
                -reactions.Fx[k, :].sum(), -reactions.Fy[k, :].sum(), -reactions.Fz[k, :].sum()
            ]
            outputs["base_M"][:, k] = np.r_[
                -reactions.Mxx[k, :].sum(), -reactions.Myy[k, :].sum(), -reactions.Mzz[k, :].sum()
            ]

            outputs["bedplate_axial_stress"][:, k] = (np.abs(Fx) / Ax + np.abs(Myy) / Sy + np.abs(Mzz) / Sz)[
                : (2 * n - 2)
            ]
            outputs["bedplate_shear_stress"][:, k] = (2.0 * (np.abs(Vy) / Asy + np.abs(Vz) / Asz) + np.abs(Mxx) / C)[
                : (2 * n - 2)
            ]
            hoop = np.zeros(2 * n - 2)

            outputs["constr_bedplate_vonmises"][:, k] = TubevonMisesStressUtilization(
                outputs["bedplate_axial_stress"][:, k],
                hoop,
                outputs["bedplate_shear_stress"][:, k],
                gamma,
                sigma_y,
            )

        # Evaluate bearing limits
        outputs["constr_mb1_defl"] = outputs["mb1_angle"] / inputs["mb1_max_defl_ang"]
        outputs["constr_mb2_defl"] = outputs["mb2_angle"] / inputs["mb2_max_defl_ang"]
        outputs["stator_deflection"] = bedplate_deflection.max()
        outputs["stator_angle"] = bedplate_angle.max()
        outputs["constr_stator_deflection"] = gamma * outputs["stator_deflection"] / stator_defl_allow
        outputs["constr_stator_angle"] = gamma * outputs["stator_angle"] / stator_angle_allow



"""WISDEM Monopile API"""

__author__ = ["Jake Nunemaker"]
__copyright__ = "Copyright 2020, National Renewable Energy Laboratory"
__maintainer__ = "Jake Nunemaker"
__email__ = "jake.nunemaker@nrel.gov"


import openmdao.api as om

from wisdem.orbit import ProjectManager


class Orbit(om.Group):
    def initialize(self):
        self.options.declare("floating", default=False)
        self.options.declare("jacket", default=False)
        self.options.declare("jacket_legs", default=0)

    def setup(self):
        # Define all input variables from all models
        self.set_input_defaults("wtiv", "example_wtiv")
        self.set_input_defaults("feeder", "example_feeder")
        self.set_input_defaults("num_feeders", 1)
        self.set_input_defaults("num_towing", 1)
        self.set_input_defaults("num_station_keeping", 3)
        self.set_input_defaults("oss_install_vessel", "example_heavy_lift_vessel")
        self.set_input_defaults("site_distance", 40.0, units="km")
        self.set_input_defaults("site_distance_to_landfall", 40.0, units="km")
        self.set_input_defaults("interconnection_distance", 40.0, units="km")
        self.set_input_defaults("plant_turbine_spacing", 7)
        self.set_input_defaults("plant_row_spacing", 7)
        self.set_input_defaults("plant_substation_distance", 1, units="km")
        self.set_input_defaults("num_port_cranes", 1)
        self.set_input_defaults("num_assembly_lines", 1)
        self.set_input_defaults("takt_time", 170.0, units="h")
        self.set_input_defaults("port_cost_per_month", 2e6, units="USD/mo")
        self.set_input_defaults("commissioning_pct", 0.01)
        self.set_input_defaults("decommissioning_pct", 0.15)
        self.set_input_defaults("site_auction_price", 100e6, units="USD")
        self.set_input_defaults("site_assessment_plan_cost", 1e6, units="USD")
        self.set_input_defaults("site_assessment_cost", 25e6, units="USD")
        self.set_input_defaults("construction_operations_plan_cost", 2.5e6, units="USD")
        self.set_input_defaults("design_install_plan_cost", 2.5e6, units="USD")
        self.set_input_defaults("boem_review_cost", 0.0, units="USD")

        self.add_subsystem(
            "orbit",
            OrbitWisdem(
                floating=self.options["floating"],
                jacket=self.options["jacket"],
                jacket_legs=self.options["jacket_legs"],
            ),
            promotes=["*"],
        )


class OrbitWisdem(om.ExplicitComponent):
    """ORBIT-WISDEM Fixed Substructure API"""

    def initialize(self):
        self.options.declare("floating", default=False)
        self.options.declare("jacket", default=False)
        self.options.declare("jacket_legs", default=0)

    def setup(self):
        """"""
        # Inputs
        # self.add_discrete_input('weather_file', 'block_island', desc='Weather file to use for installation times.')

        # Vessels
        self.add_discrete_input(
            "wtiv", "example_wtiv", desc="Vessel configuration to use for installation of foundations and turbines."
        )
        self.add_discrete_input(
            "feeder", "future_feeder", desc="Vessel configuration to use for (optional) feeder barges."
        )
        self.add_discrete_input(
            "num_feeders", 1, desc="Number of feeder barges to use for installation of foundations and turbines."
        )
        self.add_discrete_input(
            "num_towing",
            1,
            desc="Number of towing vessels to use for floating platforms that are assembled at port (with or without the turbine).",
        )
        self.add_discrete_input(
            "num_station_keeping",
            3,
            desc="Number of station keeping vessels that attach to floating platforms under tow-out.",
        )
        self.add_discrete_input(
            "oss_install_vessel",
            "example_heavy_lift_vessel",
            desc="Vessel configuration to use for installation of offshore substations.",
        )

        # Site
        self.add_input("site_depth", 40.0, units="m", desc="Site depth.")
        self.add_input("site_distance", 40.0, units="km", desc="Distance from site to installation port.")
        self.add_input(
            "site_distance_to_landfall", 50.0, units="km", desc="Distance from site to landfall for export cable."
        )
        self.add_input("interconnection_distance", 3.0, units="km", desc="Distance from landfall to interconnection.")
        self.add_input("site_mean_windspeed", 9.0, units="m/s", desc="Mean windspeed of the site.")

        # Plant
        self.add_discrete_input("number_of_turbines", 60, desc="Number of turbines.")
        self.add_input("plant_turbine_spacing", 7, desc="Turbine spacing in rotor diameters.")
        self.add_input("plant_row_spacing", 7, desc="Row spacing in rotor diameters. Not used in ring layouts.")
        self.add_input(
            "plant_substation_distance", 1, units="km", desc="Distance from first turbine in string to substation."
        )

        # Turbine
        self.add_input("turbine_rating", 8.0, units="MW", desc="Rated capacity of a turbine.")
        self.add_input("turbine_rated_windspeed", 11.0, units="m/s", desc="Rated windspeed of the turbine.")
        self.add_input("turbine_capex", 1100, units="USD/kW", desc="Turbine CAPEX")
        self.add_input("hub_height", 100.0, units="m", desc="Turbine hub height.")
        self.add_input("turbine_rotor_diameter", 130, units="m", desc="Turbine rotor diameter.")
        self.add_input("tower_mass", 400.0, units="t", desc="mass of the total tower.")
        self.add_input("tower_length", 100.0, units="m", desc="Total length of the tower.")
        self.add_input(
            "tower_deck_space",
            25.0,
            units="m**2",
            desc="Deck space required to transport the tower. Defaults to 0 in order to not be a constraint on installation.",
        )
        self.add_input("nacelle_mass", 500.0, units="t", desc="mass of the rotor nacelle assembly (RNA).")
        self.add_input(
            "nacelle_deck_space",
            25.0,
            units="m**2",
            desc="Deck space required to transport the rotor nacelle assembly (RNA). Defaults to 0 in order to not be a constraint on installation.",
        )
        self.add_discrete_input("number_of_blades", 3, desc="Number of blades per turbine.")
        self.add_input("blade_mass", 50.0, units="t", desc="mass of an individual blade.")
        self.add_input(
            "blade_deck_space",
            100.0,
            units="m**2",
            desc="Deck space required to transport a blade. Defaults to 0 in order to not be a constraint on installation.",
        )

        # Mooring
        self.add_discrete_input("num_mooring_lines", 3, desc="Number of mooring lines per platform.")
        self.add_input("mooring_line_mass", 1e4, units="kg", desc="Total mass of a mooring line")
        self.add_input("mooring_line_diameter", 0.1, units="m", desc="Cross-sectional diameter of a mooring line")
        self.add_input("mooring_line_length", 1e3, units="m", desc="Unstretched mooring line length")
        self.add_input("anchor_mass", 1e4, units="kg", desc="Total mass of an anchor")
        self.add_input("mooring_line_cost", 0.5e6, units="USD", desc="Mooring line unit cost.")
        self.add_input("mooring_anchor_cost", 0.1e6, units="USD", desc="Mooring line unit cost.")
        self.add_discrete_input("anchor_type", "drag_embedment", desc="Number of mooring lines per platform.")

        # Port
        self.add_input("port_cost_per_month", 2e6, units="USD/mo", desc="Monthly port costs.")
        self.add_input(
            "takt_time", 170.0, units="h", desc="Substructure assembly cycle time when doing assembly at the port."
        )
        self.add_discrete_input(
            "num_assembly_lines", 1, desc="Number of assembly lines used when assembly occurs at the port."
        )
        self.add_discrete_input(
            "num_port_cranes",
            1,
            desc="Number of cranes used at the port to load feeders / WTIVS when assembly occurs on-site or assembly cranes when assembling at port.",
        )

        # Floating Substructures
        self.add_input("floating_substructure_cost", 10e6, units="USD", desc="Floating substructure unit cost.")

        # Monopile
        self.add_input("monopile_length", 100.0, units="m", desc="Length of monopile (including pile).")
        self.add_input("monopile_diameter", 7.0, units="m", desc="Diameter of monopile.")
        self.add_input("monopile_mass", 900.0, units="t", desc="mass of an individual monopile.")
        self.add_input("monopile_cost", 4e6, units="USD", desc="Monopile unit cost.")

        # Jacket
        self.add_input("jacket_length", 65.0, units="m", desc="Length/height of jacket (including pile/buckets).")
        self.add_input("jacket_mass", 900.0, units="t", desc="mass of an individual jacket.")
        self.add_input("jacket_cost", 4e6, units="USD", desc="Jacket unit cost.")
        self.add_input("jacket_r_foot", 10.0, units="m", desc="Radius of jacket legs at base from centeroid.")

        # Generic fixed-bottom
        self.add_input("transition_piece_mass", 250.0, units="t", desc="mass of an individual transition piece.")
        self.add_input(
            "transition_piece_deck_space",
            25.0,
            units="m**2",
            desc="Deck space required to transport a transition piece. Defaults to 0 in order to not be a constraint on installation.",
        )
        self.add_input("transition_piece_cost", 1.5e6, units="USD", desc="Transition piece unit cost.")

        # Project
        self.add_input("site_auction_price", 100e6, units="USD", desc="Cost to secure site lease")
        self.add_input(
            "site_assessment_plan_cost", 1e6, units="USD", desc="Cost to do engineering plan for site assessment"
        )
        self.add_input("site_assessment_cost", 25e6, units="USD", desc="Cost to execute site assessment")
        self.add_input("construction_operations_plan_cost", 2.5e6, units="USD", desc="Cost to do construction planning")
        self.add_input(
            "boem_review_cost",
            0.0,
            units="USD",
            desc="Cost for additional review by U.S. Dept of Interior Bureau of Ocean Energy Management (BOEM)",
        )
        self.add_input("design_install_plan_cost", 2.5e6, units="USD", desc="Cost to do installation planning")

        # Other
        self.add_input("commissioning_pct", 0.01, desc="Commissioning percent.")
        self.add_input("decommissioning_pct", 0.15, desc="Decommissioning percent.")

        # Outputs
        # Totals
        self.add_output(
            "bos_capex", 0.0, units="USD", desc="Total BOS CAPEX not including commissioning or decommissioning."
        )
        self.add_output(
            "total_capex", 0.0, units="USD", desc="Total BOS CAPEX including commissioning and decommissioning."
        )
        self.add_output(
            "total_capex_kW", 0.0, units="USD/kW", desc="Total BOS CAPEX including commissioning and decommissioning."
        )
        self.add_output("installation_time", 0.0, units="h", desc="Total balance of system installation time.")
        self.add_output("installation_capex", 0.0, units="USD", desc="Total balance of system installation cost.")

    def compile_orbit_config_file(self, inputs, outputs, discrete_inputs, discrete_outputs):
        """"""

        floating_flag = self.options["floating"]
        jacket_flag = self.options["jacket"]

        config = {
            # Vessels
            "wtiv": "floating_heavy_lift_vessel" if floating_flag else discrete_inputs["wtiv"],
            "array_cable_install_vessel": "example_cable_lay_vessel",
            "array_cable_bury_vessel": "example_cable_lay_vessel",
            "export_cable_install_vessel": "example_cable_lay_vessel",
            "export_cable_bury_vessel": "example_cable_lay_vessel",
            # Site/plant
            "site": {
                "depth": float(inputs["site_depth"]),
                "distance": float(inputs["site_distance"]),
                "distance_to_landfall": float(inputs["site_distance_to_landfall"]),
                "mean_windspeed": float(inputs["site_mean_windspeed"]),
            },
            "landfall": {
                "interconnection_distance": float(inputs["interconnection_distance"]),
            },
            "plant": {
                "layout": "grid",
                "num_turbines": int(discrete_inputs["number_of_turbines"]),
                "row_spacing": float(inputs["plant_row_spacing"]),
                "turbine_spacing": float(inputs["plant_turbine_spacing"]),
                "substation_distance": float(inputs["plant_substation_distance"]),
            },
            # Turbine + components
            "turbine": {
                "hub_height": float(inputs["hub_height"]),
                "rotor_diameter": float(inputs["turbine_rotor_diameter"]),
                "turbine_rating": float(inputs["turbine_rating"]),
                "rated_windspeed": float(inputs["turbine_rated_windspeed"]),
                "tower": {
                    "type": "Tower",
                    "deck_space": float(inputs["tower_deck_space"]),
                    "mass": float(inputs["tower_mass"]),
                    "length": float(inputs["tower_length"]),
                },
                "nacelle": {
                    "type": "Nacelle",
                    "deck_space": float(inputs["nacelle_deck_space"]),
                    "mass": float(inputs["nacelle_mass"]),
                },
                "blade": {
                    "type": "Blade",
                    "number": int(discrete_inputs["number_of_blades"]),
                    "deck_space": float(inputs["blade_deck_space"]),
                    "mass": float(inputs["blade_mass"]),
                },
            },
            # Substructure components
            "transition_piece": {
                "type": "Transition Piece",
                "deck_space": float(inputs["transition_piece_deck_space"]),
                "mass": float(inputs["transition_piece_mass"]),
                "unit_cost": float(inputs["transition_piece_cost"]),
            },
            # Electrical
            "array_system_design": {
                "cables": ["XLPE_630mm_66kV", "XLPE_185mm_66kV"],
            },
            "export_system_design": {
                "cables": "XLPE_1000m_220kV",
                "interconnection_distance": float(inputs["interconnection_distance"]),
                "percent_added_length": 0.1,
            },
            # Phase Specific
            "OffshoreSubstationInstallation": {
                "oss_install_vessel": "floating_heavy_lift_vessel" if floating_flag else "example_heavy_lift_vessel",
                "feeder": "floating_barge" if floating_flag else "future_feeder",
                "num_feeders": int(discrete_inputs["num_feeders"]),
            },
            # Project development costs
            "project_development": {
                "site_auction_price": float(inputs["site_auction_price"]),  # 100e6,
                "site_assessment_plan_cost": float(inputs["site_assessment_plan_cost"]),  # 1e6,
                "site_assessment_cost": float(inputs["site_assessment_cost"]),  # 25e6,
                "construction_operations_plan_cost": float(inputs["construction_operations_plan_cost"]),  # 2.5e6,
                "boem_review_cost": float(inputs["boem_review_cost"]),  # 0,
                "design_install_plan_cost": float(inputs["design_install_plan_cost"]),  # 2.5e6
            },
            # Other
            "commissioning": float(inputs["commissioning_pct"]),
            "decomissioning": float(inputs["decommissioning_pct"]),
            "turbine_capex": float(inputs["turbine_capex"]),
            # Phases
            # Putting monopile or semisub here would override the inputs we assume to get from WISDEM
            "design_phases": [
                #'MonopileDesign',
                #'SemiSubmersibleDesign',
                #'MooringSystemDesign',
                #'ScourProtectionDesign',
                "ArraySystemDesign",
                "ExportSystemDesign",
                "OffshoreSubstationDesign",
            ],
        }

        # Unique design phases
        if floating_flag:
            config["install_phases"] = {
                "ExportCableInstallation": 0,
                "OffshoreSubstationInstallation": 0,
                "MooringSystemInstallation": 0,
                "MooredSubInstallation": ("MooringSystemInstallation", 0.25),
                "ArrayCableInstallation": ("MooredSubInstallation", 0.25),
            }
        else:
            fixedStr = "JacketInstallation" if jacket_flag else "MonopileInstallation"

            if jacket_flag:
                monopile = config.get("monopile", {})
                monopile["diameter"] = 10
                config["monopile"] = monopile

            config["design_phases"] += ["ScourProtectionDesign"]
            config["install_phases"] = {
                "ExportCableInstallation": 0,
                "OffshoreSubstationInstallation": 0,
                "ArrayCableInstallation": 0,
                fixedStr: 0,
                "ScourProtectionInstallation": 0,
                "TurbineInstallation": (fixedStr, 0.25),
            }

        # Unique vessels
        if floating_flag:
            vessels = {
                "support_vessel": "example_support_vessel",
                "towing_vessel": "example_towing_vessel",
                "mooring_install_vessel": "example_support_vessel",
                "towing_vessel_groups": {
                    "towing_vessels": int(discrete_inputs["num_towing"]),
                    "station_keeping_vessels": int(discrete_inputs["num_station_keeping"]),
                },
            }
        else:
            vessels = {
                "feeder": discrete_inputs["feeder"],
                "num_feeders": int(discrete_inputs["num_feeders"]),
                "spi_vessel": "example_scour_protection_vessel",
            }
        config.update(vessels)

        # Unique support structure design/assembly
        if floating_flag:
            config["port"] = {
                "sub_assembly_lines": int(discrete_inputs["num_assembly_lines"]),
                "turbine_assembly_cranes": int(discrete_inputs["num_port_cranes"]),
                "monthly_rate": float(inputs["port_cost_per_month"]),
            }

            config["substructure"] = {
                "takt_time": float(inputs["takt_time"]),
                "unit_cost": float(inputs["floating_substructure_cost"]),
                "towing_speed": 6,  # km/h
            }

            anchorstr_in = discrete_inputs["anchor_type"].lower()
            if anchorstr_in.find("drag") >= 0:
                anchorstr = "Drag Embedment"
            elif anchorstr_in.find("suction") >= 0:
                anchorstr = "Suction Pile"

            config["mooring_system"] = {
                "num_lines": int(discrete_inputs["num_mooring_lines"]),
                "line_mass": 1e-3 * float(inputs["mooring_line_mass"]),
                "line_diam": float(inputs["mooring_line_diameter"]),
                "line_length": float(inputs["mooring_line_length"]),
                "line_cost": float(inputs["mooring_line_cost"]),
                "anchor_mass": 1e-3 * float(inputs["anchor_mass"]),
                "anchor_type": anchorstr,
                "anchor_cost": float(inputs["mooring_anchor_cost"]),
            }
        else:
            config["port"] = {
                "num_cranes": int(discrete_inputs["num_port_cranes"]),
                "monthly_rate": float(inputs["port_cost_per_month"]),
            }

            config["scour_protection_design"] = {
                "cost_per_tonne": 20,
            }

            if jacket_flag:
                config["jacket"] = {
                    "type": "Jacket",
                    "height": float(inputs["jacket_length"]),
                    "num_legs": int(self.options["jacket_legs"]),
                    "deck_space": 4 * float(inputs["jacket_r_foot"]) ** 2,
                    "mass": float(inputs["jacket_mass"]),
                    "unit_cost": float(inputs["jacket_cost"]),
                }
            else:
                config["monopile"] = {
                    "type": "Monopile",
                    "length": float(inputs["monopile_length"]),
                    "diameter": float(inputs["monopile_diameter"]),
                    "deck_space": 0.25 * float(inputs["monopile_diameter"] * inputs["monopile_length"]),
                    "mass": float(inputs["monopile_mass"]),
                    "unit_cost": float(inputs["monopile_cost"]),
                }

        self._orbit_config = config
        return config

    def compute(self, inputs, outputs, discrete_inputs, discrete_outputs):
        config = self.compile_orbit_config_file(inputs, outputs, discrete_inputs, discrete_outputs)

        # Debug only
        #import json

        #fname = "orbit_dump.json"
        #with open(fname, "w") as f:
        #    f.write(json.dumps(config))

        #with open(fname) as f:
        #    config = json.loads(f.read())
        
        project = ProjectManager(config)
        project.run()

        outputs["bos_capex"] = project.bos_capex
        outputs["total_capex"] = project.total_capex
        outputs["total_capex_kW"] = project.total_capex_per_kw
        outputs["installation_time"] = project.installation_time
        outputs["installation_capex"] = project.installation_capex
